import itertools
from typing import List, Tuple, Any, Sequence, Union
import numpy

# from numpy.typing import ArrayLike
from scipy.ndimage import gaussian_filter

from aydin.it.transforms.base import ImageTransformBase
from aydin.util.log.log import lsection, lprint


class FixedPatternTransform(ImageTransformBase):
    """Fixed Axis-Aligned Pattern Suppression

    Suppresses fixed, axis aligned, offset patterns along any combination of
    axis. Given a list of lists of axis that defines axis-aligned volumes,
    intensity fluctuations of these volumes are stabilised. You can suppress
    intensity fluctuation over time, suppress fixed offsets per pixel over
    time, suppress intensity fluctuations per row, per column, and more...

    For example, assume an image with dimensions tyx (t+2D), and you want to
    suppress fluctuations of intensity along the t axis, then you provide:
    axes=[[0]] (or simply 0 or [0]) which means that the average intensity
    for all planes along t (axis=0) will be stabilised. If instead you want
    to suppress some fixed background offset over xy planes, then you do:
    axes=[[1,2]]. If you want to do both, then you use: axes=[[0], [1,
    2]]. Please note that these corrections are applied in the order
    specified by the list of axis combinations. It is not recommended to
    reapply the pattern after denoising, unless the pattern itself is of
    value and is not considered noise.
    """

    preprocess_description = (
        "Axis-aligned pattern suppression" + ImageTransformBase.preprocess_description
    )
    postprocess_description = (
        "Reapplies pattern" + ImageTransformBase.postprocess_description
    )
    postprocess_supported = True
    postprocess_recommended = False

    def __init__(
        self,
        axes: Union[None, int, Sequence[int], List[Sequence[int]]] = None,
        percentile: float = 1,
        sigma: float = 0.5,
        priority: float = 0.09,
        **kwargs,
    ):

        """
        Constructs a Background Correction

        Parameters
        ----------
        axes : Union[None, int, Sequence[int], List[Sequence[int]]]
            List of axis combinations. The order provided
            is the order in which the corrections are applied.
            If None the axes are automatically determined.
        percentile : float
            Percentile value used for estimating brightness.
        sigma : float
            Sigma of the Gaussian filter applied on the
            detected pattern. The higher this value the
            less high-frequency fluctuations are corrected.
        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.
        """
        super().__init__(priority=priority, **kwargs)

        # In case a single integer is passed:
        if axes is not None and type(axes) == int:
            axes = list((axes,))

        # In case only one axis combination is given:
        if axes is not None and len(axes) > 0 and type(axes[0]) == int:
            axes = list((axes,))

        # normalise to tuple:
        axes = (
            [tuple(ac) for ac in axes] if axes is not None and len(axes) > 0 else axes
        )

        self.axis_combinations = axes
        self.percentile = percentile
        self.sigma = sigma
        self._corrections = {}

        lprint(f"Instanciating: {self}")

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_corrections']
        return state

    def __str__(self):
        return (
            f'{type(self).__name__}'
            f' (percentile={self.percentile},'
            f' sigma={self.sigma})'
        )

    def __repr__(self):
        return self.__str__()

    def preprocess(self, array):

        with lsection(
            f"Removing axis-aligned fixed patterns for array of shape: {array.shape} and dtype: {array.dtype}:"
        ):
            self._original_dtype = array.dtype
            new_array = array.astype(dtype=numpy.float32, copy=True)

            overall_value = numpy.percentile(
                new_array, q=self.percentile, keepdims=True
            )

            if self.axis_combinations is None:
                # Default:
                axis_combinations = _all_axis_combinations(array.ndim)
            else:
                # Invert the meaning of the axis:
                axis_combinations = (
                    tuple((a for a in range(array.ndim) if a not in ac))
                    for ac in self.axis_combinations
                )

            self._axis_combinations = axis_combinations
            self._corrections = {}

            for axis_combination in axis_combinations:
                lprint(f"Suppressing variations across hyperplane: {axis_combination}")
                value = numpy.percentile(
                    new_array, q=self.percentile, axis=axis_combination, keepdims=True
                )
                value = gaussian_filter(value, sigma=self.sigma)

                correction = overall_value - value
                new_array += correction

                self._corrections[axis_combination] = correction

            self.overall_value = overall_value

            return new_array

    def postprocess(self, array):

        if not self.do_postprocess:
            return array

        with lsection(
            f"Adding back axis-aligned fixed pattern to array of shape: {array.shape} and dtype: {array.dtype}:"
        ):

            # Allocate new array to store result:
            new_array = array.astype(dtype=numpy.float32, copy=True)

            for axis_combination in reversed(self._axis_combinations):
                correction = self._corrections[axis_combination]
                new_array -= correction
                self._corrections[axis_combination] = correction

            new_array = new_array.astype(self._original_dtype, copy=False)

            return new_array


def _axis_combinations(ndim: int, n: int) -> List[Tuple[Any, ...]]:
    return list(itertools.combinations(range(ndim), n))


def _all_axis_combinations(ndim: int):
    axis_combinations = []
    for dim in range(1, ndim):
        combinations = _axis_combinations(ndim, dim)
        axis_combinations.extend(combinations)
    return axis_combinations



import numpy
from skimage.util import random_noise

from aydin.io.datasets import camera, normalise, pollen
from aydin.it.transforms.motion import (
    MotionStabilisationTransform,
    _find_shift,
    _measure_shifts,
    _shift_transform,
)


def test_phase_correlation():
    a = camera()
    b = numpy.roll(camera(), shift=(30, 50), axis=tuple(range(a.ndim)))

    # phase_correlation = _phase_correlation(image=b, reference_image=a)
    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(a, name='a')
    #     viewer.add_image(b, name='b')
    #     viewer.add_image(phase_correlation, name='phase_correlation')

    shift, _ = _find_shift(b, a)

    # print(shift)
    assert (numpy.abs(shift - [30, 50]) < 0.1).all()


def test_measure_shifts_and_transform():
    shifts = tuple((5 * i, int(0.5 * i * i)) for i in range(10))

    # print('')
    # pprint(shifts)

    image = normalise(camera())
    array = numpy.stack(
        [numpy.roll(image, shift=shift, axis=(0, 1)) for shift in shifts]
    )

    measured_shifts, _ = _measure_shifts(array, reference_index=0)

    # pprint(measured_shifts)

    for s, ms in zip(shifts, measured_shifts):
        assert (numpy.array(s) == numpy.array(ms)).all()

    measured_shifts, _ = _measure_shifts(array, reference_index=0, center=False)
    # def _shift_transform(array, shifts, pad, crop, pad_mode='wrap', inverse=False):
    motion_corrected_array = _shift_transform(
        array.copy(), -measured_shifts, pad=False, crop=False
    )

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(array, name='array')
    #     viewer.add_image(motion_corrected_array, name='motion_corrected_array')

    assert (motion_corrected_array == numpy.stack([image for _ in shifts])).all()


def test_correct_uncorrect():
    shifts = tuple((5 * i, int(0.5 * i * i)) for i in range(10))

    # print('')
    # pprint(shifts)

    image = normalise(pollen())[0:256, 0:256]
    array = numpy.stack(
        [add_noise(numpy.roll(image, shift=shift, axis=(0, 1))) for shift in shifts]
    )

    mc = MotionStabilisationTransform(axes=0)

    corrected_array = mc.preprocess(array.copy())
    uncorrected_array = mc.postprocess(corrected_array.copy())

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(array, name='array')
    #     viewer.add_image(corrected_array, name='corrected_array')
    #     viewer.add_image(uncorrected_array, name='uncorrected_array')

    assert array.dtype == uncorrected_array.dtype
    assert (array == uncorrected_array).all()


def add_noise(image, intensity=4, variance=0.4):
    noisy = image
    if intensity is not None:
        noisy = numpy.random.poisson(image * intensity) / intensity
    noisy = random_noise(noisy, mode="gaussian", var=variance)
    noisy = noisy.astype(numpy.float32)
    return noisy



import re


def break_text(
    text, max_width: int = 80, clear_existing_breaks: bool = True, line_break='\n'
):

    # We remove existing line breaks except double line breaks if requested:
    if clear_existing_breaks:
        text = text.replace('\n\n', '<!LB!>')
        text = text.replace('\n', ' ')
        text = text.replace('<!LB!>', '\n\n')

    # remove repeated spaces:
    text = re.sub(' +', ' ', text)

    # string_no_punctuation = re.sub("[^\w\s]", "", text)
    words = re.split(r'(\s+)', text)

    # words = re.findall(rf'[\w\W\-.",]+', text)

    output_text = ''
    counter = 0
    for word in words:
        if counter >= max_width or '\n' in word:
            output_text += line_break
            counter = 0

        if not (word == ' ' and counter == 0):
            if '\n' in word:
                word = word.strip() + '\n'
            output_text += word
            counter += len(_clean_html(word))

    return output_text


_html_clean_re = re.compile('<.*?>')


def _clean_html(text):
    if len(text.strip()) == 0:
        return text
    else:
        clean_text = text.replace('<a', '')
        clean_text = clean_text.replace('href=', '')
        clean_text = re.sub(r'https?:\/\/.*[\r\n]*', '', clean_text, flags=re.MULTILINE)
        clean_text = re.sub(_html_clean_re, '', clean_text)
        return clean_text



#!/usr/bin/env python
"""
Test focal scan G-L PSF.
"""
import numpy

from aydin.util.psf.microscope_psf import MicroscopePSF


def test_01():
    """
    Particle on surface.
    """
    psf = MicroscopePSF()
    rv = numpy.arange(0.0, 1.01, 0.1)
    zv = numpy.arange(-1.0, 1.01, 0.2)

    fast_rz = psf.gLZRFocalScan(rv, zv)
    slow_rz = psf.gLZRFocalScanSlow(rv, zv)

    assert numpy.allclose(fast_rz, slow_rz)


def test_02():
    """
    Particle above surface.
    """
    psf = MicroscopePSF()
    rv = numpy.arange(0.0, 1.01, 0.1)
    zv = numpy.arange(-1.0, 1.01, 0.2)

    fast_rz = psf.gLZRFocalScan(rv, zv, pz=0.5)
    slow_rz = psf.gLZRFocalScanSlow(rv, zv, pz=0.5)

    assert numpy.allclose(fast_rz, slow_rz, atol=1.0e-4, rtol=1.0e-4)


def test_03():
    """
    Detector offset.
    """
    psf = MicroscopePSF()
    rv = numpy.arange(0.0, 1.01, 0.1)
    zv = numpy.arange(-1.0, 1.01, 0.2)

    zd = psf.parameters["zd0"] + 1000
    fast_rz = psf.gLZRFocalScan(rv, zv, zd=zd)
    slow_rz = psf.gLZRFocalScanSlow(rv, zv, zd=zd)

    assert numpy.allclose(fast_rz, slow_rz)


def test_04():
    """
    Particle scan.
    """
    psf = MicroscopePSF()
    rv = numpy.arange(0.0, 1.01, 0.1)
    pv = numpy.arange(0.0, 2.01, 0.1)

    fast_rz = psf.gLZRParticleScan(rv, pv)
    slow_rz = psf.gLZRParticleScanSlow(rv, pv)

    assert numpy.allclose(fast_rz, slow_rz, rtol=1.0e-4, atol=1.0e-4)


def test_05():
    """
    Particle scan, focus offset.
    """
    psf = MicroscopePSF()
    rv = numpy.arange(0.0, 1.01, 0.1)
    pv = numpy.arange(1.0, 3.01, 0.2)

    fast_rz = psf.gLZRParticleScan(rv, pv, zv=-2.0)
    slow_rz = psf.gLZRParticleScanSlow(rv, pv, zv=-2.0)

    assert numpy.allclose(fast_rz, slow_rz, rtol=1.0e-3, atol=1.0e-3)



from functools import partial
import numpy
from numpy import random

from aydin.it.classic_denoisers.butterworth import denoise_butterworth
from aydin.util.j_invariance.j_invariant_classic import calibrate_denoiser_classic
from aydin.util.crop.rep_crop import representative_crop


def resolution_estimate(image, precision: float = 0.01, display_images: bool = False):
    """Estimation of isotropic resolution in normalised frequency (within [0, 1]).
    Work best

    Parameters
    ----------
    image : numpy.typing.ArrayLike
        Image to estimate (isotropic) resolution from
    precision: float
        Precision of the search
    display_images: bool
        Display image, for debugging purposes.

    Returns
    -------
    Estimate of resolution in normalised frequency -- which assumes that the
    sampling frequency is 1. Values are between 0 and 1, with 0 meaning no
    resolution at all, and 1 means full resolution.
    Also returns the crop used for the estimate.

    """

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=128000, equal_sides=True)

    # Convert image to float if needed:
    crop = crop.astype(dtype=numpy.float32, copy=False)

    # We need to add noise for this to work:
    sigma = 0.01 * numpy.std(crop)
    crop += random.normal(scale=sigma, size=crop.shape)

    # ranges:
    freq_cutoff_range = numpy.arange(0.001, 1.0, precision)

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {'freq_cutoff': freq_cutoff_range}

    # Partial function:
    _denoise_sobolev = partial(denoise_butterworth, multi_core=False, order=2)

    # Calibrate denoiser
    best_parameters = calibrate_denoiser_classic(
        crop,
        _denoise_sobolev,
        denoise_parameters=parameter_ranges,
        display_images=display_images,
    )

    norm_frequency = best_parameters.pop('freq_cutoff')

    return norm_frequency, crop



# flake8: noqa
# flake8: noqa

import numpy
from aydin.it.pytorch.lucyrichardson.clipped_lr import richardson_lucy_shrink
from scipy import ndimage
from scipy.fft import dct
from scipy.signal import convolve2d
from skimage.exposure import rescale_intensity
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.analysis.resolution_estimate import signal_to_noise_ratio, spectrum_2d
from aydin.features.fast.fast_features import FastMultiscaleConvolutionalFeatures
from aydin.io.datasets import normalise, add_noise, newyork
from aydin.it.it_classic import ImageTranslatorClassic
from aydin.regression.gbm import GBMRegressor
from aydin.util.psf.simple_microscope_psf import SimpleMicroscopePSF


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


def dctpsnr(true_image, test_image):
    true_image = n(dct(dct(true_image, axis=0), axis=1))
    test_image = n(dct(dct(test_image, axis=0), axis=1))

    return psnr(true_image, test_image)


deconv = richardson_lucy_shrink

image = newyork()
image = normalise(image.astype(numpy.float32))

psf = SimpleMicroscopePSF()
psf_xyz_array = psf.generate_xyz_psf(dxy=0.406, dz=0.406, xy_size=17, z_size=17)
print(psf_xyz_array.shape)
kernel_psf = psf_xyz_array[0]

blurred_image = convolve2d(image, kernel_psf, mode='same', boundary='symm')
noisy_and_blurred_image = add_noise(blurred_image, intensity=100, variance=0.001)

# noisy = numpy.stack([add_noise(blurred_image, intensity=1000, variance=0.01*i) for i in range(30)])
# noisy_spectrum = numpy.stack([spectrum(i, log=True)[0] for i in noisy])
# snr_values = [signal_to_noise_ratio(i) for i in noisy]
# print(snr_values)
#
# import napari
#
# with napari.gui_qt():
#     viewer = napari.Viewer()
#     viewer.add_image(noisy, name='noisy')
#     viewer.add_image(noisy_spectrum, name='noisy_spectrum')


it = ImageTranslatorClassic(
    feature_generator=FastMultiscaleConvolutionalFeatures(exclude_scale_one=False),
    regressor=GBMRegressor(n_estimators=2000),
    normaliser_type='identity',
)

# Try denoising first:
it.train(noisy_and_blurred_image)
noisy_and_blurred_image = it.translate(noisy_and_blurred_image)

deconvolved_image = numpy.stack(
    [
        deconv(noisy_and_blurred_image, kernel_psf, iterations=i)
        for i in range(0, 60, 30)
    ]
)

supervised_ssim = [ssim(i, image) for i in deconvolved_image]
print(supervised_ssim)
print(supervised_ssim.index(max(supervised_ssim)))

supervised_psnr = [psnr(i, image) for i in deconvolved_image]
print(supervised_psnr)
print(supervised_psnr.index(max(supervised_psnr)))

input_image = noisy_and_blurred_image
mask = numpy.random.rand(*image.shape) < 0.001
masked_input_image = (
    input_image * ~mask
    + ndimage.median_filter(
        input_image, footprint=numpy.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]])
    )
    * mask
)

deconvolved_image = numpy.stack(
    [deconv(masked_input_image, kernel_psf, iterations=i) for i in range(0, 60, 6)]
)

reconvolved_image = numpy.stack(
    [convolve2d(i, kernel_psf, mode='same', boundary='symm') for i in deconvolved_image]
)

diff_image = numpy.stack([abs(i - input_image) for i in reconvolved_image])

deconvolved_spectrum = numpy.stack(
    [spectrum_2d(i, log=True)[0] for i in deconvolved_image]
)

similarity_to_input = list([ssim(input_image, i) for i in deconvolved_image])
ssim_indicator = list(
    [
        0.5 * ssim(input_image, d) + 0.5 * ssim(input_image, r)
        for d, r in zip(deconvolved_image, reconvolved_image)
    ]
)

print(similarity_to_input)
print(similarity_to_input.index(max(similarity_to_input)))
print(ssim_indicator)
print(ssim_indicator.index(max(ssim_indicator)))

mutual_info_deconv = list([mutual_info(input_image, i) for i in deconvolved_image])
mutual_info_reconv = list([mutual_info(input_image, i) for i in reconvolved_image])
mutual_info_indicator = list(
    [0.5 * i + 0.5 * j for i, j in zip(mutual_info_deconv, mutual_info_reconv)]
)

snr = list([signal_to_noise_ratio(i) for i in deconvolved_image])

from matplotlib import pyplot as plt

plt.plot(ssim_indicator[1:], label='ssim_indicator')
# plt.plot(mutual_info_indicator[1:], label='mutual_info_indicator')
plt.plot(similarity_to_input[1:], label='similarity')
# plt.plot(mutual_info_deconv[1:], label='mutual_info_deconv')
# plt.plot(mutual_info_reconv[1:], label='mutual_info_reconv')
plt.plot(snr[1:], label='snr')
plt.legend()
plt.show()

import napari

with napari.gui_qt():
    viewer = napari.Viewer()
    # viewer.add_image(module_deconvolved_image, name='module_deconvolved_image')
    # viewer.add_image(si_deconvolved_denoised_image, name='si_deconvolved_denoised_image')
    viewer.add_image(diff_image, name='diff_image')
    viewer.add_image(masked_input_image, name='masked_input_image')
    viewer.add_image(reconvolved_image, name='reconvolved_image')
    viewer.add_image(deconvolved_image, name='deconvolved_image')
    viewer.add_image(deconvolved_spectrum, name='deconvolved_spectrum')
    viewer.add_image(noisy_and_blurred_image, name='noisy_and_blurred_image')
    viewer.add_image(image, name='image')



from functools import partial
from typing import Optional

import numpy
from skimage.restoration import denoise_tv_bregman
from skimage.restoration._denoise import _denoise_tv_chambolle_nd

from aydin.util.crop.rep_crop import representative_crop
from aydin.util.j_invariance.j_invariant_smart import calibrate_denoiser_smart


def calibrate_denoise_tv(
    image,
    crop_size_in_voxels: Optional[int] = 64000,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the Total Variation (TV) denoiser for the given image and
    returns the optimal parameters obtained using the N2S loss.

    Note: we use the scikt-image implementation of TV denoising.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate TV denoiser for.

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate denoiser.
        (advanced)

    display_images: bool
        When True the denoised images encountered during optimisation are shown

    other_fixed_parameters: dict
        Any other fixed parameters

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.

    """
    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels)

    # Sigma spatial range:
    range = max(1e-10, numpy.max(image) - numpy.min(image))
    # weight_range = np.arange(0.01, 1.5*std, 0.05*std) ** 1.5
    weight_range = (0.01 * range, 1.5 * range)

    # Algorithms:
    algorithms = ['bregman', 'chambolle'] if image.ndim <= 2 else ['chambolle']
    # Note: scikitimage implementation of bregman does not support more than 2D

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {
        'weight': weight_range,
        'isotropic': [False, True],
        'algorithm': algorithms,
        # 'algorithm': ['chambolle','bregman'], #
    }

    # Partial function:
    _denoise_tv = partial(denoise_tv, **other_fixed_parameters)

    # Calibrate denoiser
    best_parameters = (
        calibrate_denoiser_smart(
            crop,
            _denoise_tv,
            denoise_parameters=parameter_ranges,
            display_images=display_images,
        )
        | other_fixed_parameters
    )

    # Memory needed:
    memory_needed = image.nbytes * 2  # gradient image

    return denoise_tv, best_parameters, memory_needed


def denoise_tv(image, algorithm: str = 'bregman', weight: float = 1, **kwargs):
    """
    Denoises the given image using either scikit-image
    implementation of Bregman or Chambolle <a
    href="https://en.wikipedia.org/wiki/Total_variation_denoising">Total
    Variation (TV) denoising</a>. We attempt to rescale the weight parameter to
    obtain similar results between Bregman and Chambolle for the
    same weight.
    \n\n
    Note: Because of limitations of the current scikit-image
    implementation, if an image with more than 2 dimensions
    is passed, we use the Chambolle implementation as it
    supports nD images...

    Parameters
    ----------
    image: ArrayLike
        Image to denoise

    algorithm: str
        Algorithm to apply: either 'bregman' or 'chambolle'

    weight: float
        Weight to balance data term versus prior term. Larger values correspond
        to a stronger prior during optimisation, and thus more aggressive
        denoising

    kwargs
        Any other parameters to be passed to scikit-image implementations

    Returns
    -------
    Denoised image
    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # algorithm = parameters.pop('weight')
    if algorithm == 'bregman' and image.ndim <= 2:
        return denoise_tv_bregman(image, weight=weight * 10, **kwargs)
    if algorithm == 'chambolle' or image.ndim > 2:
        if 'isotropic' in kwargs:
            kwargs.pop('isotropic')
        return _denoise_tv_chambolle_nd(image, weight=weight, **kwargs)



import os
from abc import ABC, abstractmethod
from os.path import join
from typing import Tuple
import jsonpickle
import numexpr
import numpy

from aydin.util.misc.json import encode_indent
from aydin.util.log.log import lprint


class NormaliserBase(ABC):
    """Normaliser base class

    Parameters
    ----------
    clip : bool
    epsilon : float
    shape_normalisation : bool
    transform : str

    """

    epsilon: float
    leave_as_float: bool
    clip: bool
    original_dtype: numpy.dtype

    def __init__(self, clip=True, epsilon=1e-21):
        """Constructs a normalisers

        Parameters
        ----------
        clip : bool
        epsilon : float
        """
        self.epsilon = epsilon
        self.clip = clip

        self.rmin = None
        self.rmax = None

        self.axis_permutation = None
        self.permutated_image_shape = None

    def save(self, path: str, name='default'):
        """Saves an 'all-batteries-included' normalisers at a given path (folder).

        Parameters
        ----------
        path : str
            path to save to
        name : str

        Returns
        -------
        json
            frozen
        """
        os.makedirs(path, exist_ok=True)

        frozen = encode_indent(self)

        lprint(f"Saving normalisers to: {path}")
        with open(join(path, f"normaliser_{name}.json"), "w") as json_file:
            json_file.write(frozen)

        return frozen

    @staticmethod
    def load(path: str, name='default'):
        """Returns an 'all-batteries-included' normalisers from a given path (folder).

        Parameters
        ----------
        path : str
            path to load from.
        name : str

        Returns
        -------
        str
            thawed

        """
        lprint(f"Loading normalisers from: {path}")
        with open(join(path, f"normaliser_{name}.json"), "r") as json_file:
            frozen = json_file.read()

        thawed = jsonpickle.decode(frozen)

        return thawed

    @abstractmethod
    def calibrate(self, array) -> Tuple[float, float]:
        """Calibrates this normalisers given an array.

        Parameters
        ----------
        array : numpy.ndarray
            array to use for calibration

        Returns
        -------
        array : numpy.ndarray

        """
        raise NotImplementedError()

    def normalise(self, array):
        """Normalises the given array in-place (if possible).

        Parameters
        ----------
        array : numpy.ndarray
            array to normalise
        batch_dims : list
        channel_dims : list

        Returns
        -------
        array : numpy.ndarray

        """
        array = array.astype(numpy.float32, copy=True)

        if self.rmin is not None and self.rmax is not None:
            min_value = numpy.float32(self.rmin)
            max_value = numpy.float32(self.rmax)
            epsilon = numpy.float32(self.epsilon)

            try:
                # We perform operation in-place with numexpr if possible:
                numexpr.evaluate(
                    "((array - min_value) / ( max_value - min_value + epsilon ))",
                    out=array,
                )
                if self.clip:
                    numexpr.evaluate(
                        "where(array<0,0,where(array>1,1,array))", out=array
                    )

            except ValueError:
                array -= min_value
                array /= max_value - min_value + epsilon
                if self.clip:
                    array = numpy.clip(array, 0, 1)  # , out=array

        return array

    def denormalise(
        self,
        array: numpy.ndarray,
        denormalise_values=True,
        leave_as_float=False,
        clip=True,
    ):
        """Denormalises the given array in-place (if possible).

        Parameters
        ----------
        array : numpy.ndarray
        denormalise_values : bool
        leave_as_float : bool
        clip : bool

        Returns
        -------
        array : numpy.ndarray

        """

        # we copy the array to preserve the original array:
        array = numpy.copy(array)

        if denormalise_values:
            if self.rmin is not None and self.rmax is not None:

                min_value = numpy.float32(self.rmin)
                max_value = numpy.float32(self.rmax)
                epsilon = numpy.float32(self.epsilon)

                try:
                    # We perform operation in-place with numexpr if possible:

                    if self.clip and clip:
                        numexpr.evaluate(
                            "where(array<0,0,where(array>1,1,array))",
                            out=array,
                            casting='unsafe',
                        )

                    numexpr.evaluate(
                        "array * (max_value - min_value + epsilon) + min_value ",
                        out=array,
                        casting='unsafe',
                    )

                except ValueError:
                    if self.clip and clip:
                        array = numpy.clip(array, 0, 1)  # , out=array
                    array *= max_value - min_value + epsilon
                    array += min_value

            if not leave_as_float and self.original_dtype != array.dtype:
                if numpy.issubdtype(self.original_dtype, numpy.integer):
                    # If we cast back to integer, we need to avoid overflows first!
                    type_info = numpy.iinfo(self.original_dtype)

                    if not (self.clip and clip):
                        array = array + (type_info.min - array.min())
                        array = (array * type_info.max) / array.max()

                    array = array.clip(type_info.min, type_info.max, out=array)
                array = array.astype(self.original_dtype)

        return array



from skimage.restoration import denoise_wavelet

from aydin.io.datasets import cropped_newyork
from aydin.it.classic_denoisers.demo.demo_2D_wavelet import demo_wavelet
from aydin.it.classic_denoisers.test.util_test_nd import check_nd


def test_wavelet():
    assert demo_wavelet(cropped_newyork(), display=False) >= 0.587 - 0.01


def test_wavelet_nd():
    check_nd(denoise_wavelet)



from typing import Sequence, Tuple

import numpy
from scipy.ndimage import median_filter

from aydin.features.groups.base import FeatureGroupBase
from aydin.util.log.log import lprint


class MedianFeatures(FeatureGroupBase):
    """
    Median Feature Group class
    """

    def __init__(self, radii: Sequence[int]):
        super().__init__()
        self.radii = tuple(radii)
        self.image = None
        self.excluded_voxels: Sequence[Tuple[int, ...]] = []

        self.kwargs = None

    @property
    def receptive_field_radius(self) -> int:
        radius = max(r for r in self.radii)
        return radius

    def num_features(self, ndim: int) -> int:
        return len(self.radii)

    def prepare(self, image, excluded_voxels=None, **kwargs):
        if excluded_voxels is None:
            excluded_voxels = []

        self.image = image
        self.excluded_voxels = excluded_voxels
        self.kwargs = kwargs

    def compute_feature(self, index: int, feature):
        radius = self.radii[index]
        lprint(
            f"Median feature: {index}, radius: {radius}, excluded_voxels={self.excluded_voxels}"
        )
        footprint_shape = (2 * radius + 1,) * self.image.ndim
        footprint = numpy.ones(footprint_shape)

        # We need to modify the footprint to take into account the excluded voxels:
        for excluded_voxel in self.excluded_voxels:

            # First we check if the excluded voxel falls within the footprint of the feature:
            if all(
                (
                    -s // 2 <= v <= s // 2
                    for s, v in zip(footprint.shape, excluded_voxel)
                )
            ):
                # Here is the coordinate of the excluded voxel
                coord = tuple(
                    (s // 2 + v for s, v in zip(footprint.shape, excluded_voxel))
                )

                aslice = tuple((slice(i, i + 1, None) for i in coord))
                footprint[aslice] = 0

        median_filter(self.image, footprint=footprint, output=feature)

    def finish(self):
        # Here we cleanup any resource alocated for the last feature computation:
        self.image = None
        self.excluded_voxels = None
        self.kwargs = None



# from click.testing import CliRunner
#
# from aydin.cli.cli import update
#
#
# def test_update():
#     runner = CliRunner()
#     result = runner.invoke(update)
#     assert result.exit_code == 0


# def test_noise2self():
#     runner = CliRunner()
#     result = runner.invoke(noise2self, [])
#     assert result.exit_code == 1
#
#
# def test_info():
#     runner = CliRunner()
#     result = runner.invoke(info, [])
#     assert result.exit_code == 1



import numpy
from skimage.exposure import rescale_intensity

from aydin.features.groups.dct import DCTFeatures
from aydin.io.datasets import camera


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32, copy=False), in_range='image', out_range=(0, 1)
    )


def test_dct_feature_group():
    # get image:
    image = n(camera().astype(numpy.float32))

    # Instantiates DCT features:
    convolutions = DCTFeatures(size=5, max_freq=2, power=1)
    assert convolutions.num_features(image.ndim) == 25
    assert convolutions.receptive_field_radius == 2

    # Set image:
    convolutions.prepare(image)

    # compute features and check their valididty:
    feature = numpy.empty_like(image)

    # Compute features:
    for index in range(convolutions.num_features(image.ndim)):
        convolutions.compute_feature(index=index, feature=feature)
        assert (feature != image).any()



from sklearn.svm import SVR, LinearSVR

from aydin.regression.base import RegressorBase
from aydin.util.log.log import lprint, lsection


class SVRegressor(RegressorBase):
    """Support Vector Regressor.
    \n\n
    Note: Way too slow when non-linear, nearly useless...
    When using linear much faster, but does not perform better
    than linear regression.
    """

    def __init__(self, linear: bool = True):
        """Constructs a linear regressor.

        Parameters
        ----------
        linear : bool
            Flag to choose between a linear or non-linear SVR

        """
        super().__init__()
        self.linear = linear

    def _fit(
        self, x_train, y_train, x_valid=None, y_valid=None, regressor_callback=None
    ):
        """Fits function y=f(x) given training pairs (x_train, y_train).
        Stops when performance stops improving on the test dataset: (x_test, y_test).
        """

        if self.linear:
            model = LinearSVR()  # gamma='scale')
        else:
            model = SVR(gamma='scale')

        model.fit(x_train, y_train)

        return _SVRModel(model)


class _SVRModel:
    def __init__(self, model):
        self.model = model
        self.loss_history = {'training': [], 'validation': []}

    def _save_internals(self, path: str):
        pass

    def _load_internals(self, path: str):
        pass

    def predict(self, x):
        with lsection("SVR regressor prediction"):

            lprint(f"Number of data points             : {x.shape[0]}")
            lprint(f"Number of features per data points: {x.shape[-1]}")

            with lsection("SVR prediction now"):
                prediction = self.model.predict(x)

            lprint("SVR regressor predicting done!")
            return prediction



#!/usr/bin/env python
"""
Test focal scan G-L PSF.
"""
from pprint import pprint

from aydin.util.psf.simple_microscope_psf import SimpleMicroscopePSF


def demo_xyz():
    """
    Particle scan, focus offset.
    """
    psf = SimpleMicroscopePSF()
    pprint(psf.parameters)

    psf_xyz_array = psf.generate_xyz_psf(dxy=0.406, dz=0.406, xy_size=17, z_size=17)

    print(psf_xyz_array.shape)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(psf_xyz_array, name='fast_rz')


demo_xyz()



import warnings
import functools


def deprecated(func):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.

    Parameters
    ----------
    func : callable

    Returns
    -------
    callable

    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func



import sys
from setuptools import setup


if sys.version_info < (3, 9):
    sys.stderr.write(
        f'You are using Python '
        + "{'.'.join(str(v) for v in sys.version_info[:3])}.\n\n"
        + 'aydin only supports Python 3.9 and above.\n\n'
        + 'Please install Python 3.9 using:\n'
        + '  $ pip install python==3.9\n\n'
    )
    sys.exit(1)

# with open(os.path.join('requirements', 'default.txt')) as f:
#     default_requirements = [
#         line.strip() for line in f if line and not line.startswith('#')
#     ]
#
# INSTALL_REQUIRES = default_requirements
# REQUIRES = []

# Handle pyopencl
# os.system("pip install -r " + os.path.join('requirements', 'pyopencl.txt'))

setup(
    use_scm_version={"write_to": "aydin/_version.py"},
    setup_requires=['setuptools_scm'],
    entry_points={'console_scripts': ['aydin=aydin.cli.cli:cli']},
)



from copy import deepcopy
import numpy as np
from scipy.ndimage import median_filter

from aydin.util.log.log import lprint


def masker(batch_vol, i=None, mask_shape=None, p=None):
    """
    Single mask generator.

    Parameters
    ----------
    batch_vol
        batch volume, desn't include batch and ch dimensions
    i
    mask_shape
        mask shape e.g. (3, 3)
    p
        possibility of masked pixels on random masking approach

    Returns
    -------
    mask

    Notes
    -----
    Spec.
    The core function to create mask fixed_pattern.
    Can create both checkerbox and random masks.
    The output should have the same dimension size as the that of the image to be masked.

    Input
    patch size:
    iteration index: generate different mask patterns when iteration index changes
    mask size: a unit size of mask. e.g. unit size of checkerbox
    mask ratio for random masking: how much % of the pixels should be masked in random masking method
    """

    if p:
        mask = np.random.uniform(size=batch_vol)
        mask[mask > p] = 1
        mask[mask <= p] = 0
        mask = mask.astype(bool)
    else:
        i = i % np.prod(mask_shape)
        mask = np.zeros(np.prod(mask_shape), dtype=bool)
        mask[i] = True
        mask = mask.reshape(mask_shape)
        rep = np.ceil(np.asarray(batch_vol) / np.asarray(mask_shape)).astype(int)
        mask = np.tile(mask, tuple(rep))
        ind = tuple([slice(batch_vol[i]) for i in range(len(batch_vol))])
        mask = mask[ind]
    return mask


def med_filter(input_image):
    """
    Parameters
    ----------
    input_image

    Returns
    -------
    filtered image

    Notes
    -----
    Spec.
    This median filter function is used to replace validation pixels the median values of surround pixels.
    """
    k = [1] + [3 for _ in input_image.shape[1:-1]] + [1]
    kernel = np.ones(k)
    k = [0] + [1 for _ in range(len(input_image.shape[1:-1]))] + [0]
    kernel[tuple(k)] = 0
    return median_filter(input_image, footprint=kernel)


def maskedgen(image, batch_size, mask_size, replace_by='zero'):
    """
    Mask generator. Returns a generator.

    Parameters
    ----------
    image
        the image to be masked
    batch_size
        generator will generate <mini batch size> of slices a time to feed to CNN model
    mask_size
        a unit size of checkerbox
    replace_by
        an argument to select whether replacing masked pixels with zero or median values of the surrounding pixels

    Returns
    -------
    mask generator
    """
    patch_size = image.shape[1:-1]
    img_output = deepcopy(image)

    j = 0
    num_cycle = np.ceil(img_output.shape[0] / batch_size)
    lprint(f'Masked pixels are replaced by {replace_by}')
    while True:
        i = np.mod(j, num_cycle).astype(int, copy=False)
        image_batch = img_output[batch_size * i : batch_size * (i + 1)]
        for i in range(int(np.prod(mask_size))):
            mask = masker(
                patch_size, i, mask_size, p=None
            )  # generate a 2D mask with same size of image_batch; True value is masked.
            masknega = np.broadcast_to(
                np.expand_dims(np.expand_dims(mask, 0), -1), image_batch.shape
            )  # broadcast the 2D mask to batch dimension ready for multiplication
            train_img = (
                np.broadcast_to(
                    np.expand_dims(np.expand_dims(~mask, 0), -1), image_batch.shape
                )
                * image_batch
            )  # pixels in training image are blocked by multiply by 0

            if replace_by == 'random':
                train_img = train_img + np.random.random(patch_size) * masknega
            elif replace_by == 'median':
                train_img = train_img + med_filter(image_batch) * masknega

            target_img = masknega * image_batch
            j += 1
            yield {
                'input': train_img,
                'input_msk': masknega.astype(np.float32),
            }, target_img


def randmaskgen(
    image,
    batch_size,
    p_maskedpixels=None,
    replace_by='zero',  # 'zero' or 'random' or 'median'
):
    """
    Create a random mask generator.

    Parameters
    ----------
    image
        image to be masked
    batch_size
        generator will generate <mini batch size> of slices a time to feed to CNN model
    p_maskedpixels
        how much % of the pixels should be masked in random masking method
    replace_by
        an argument to select whether replacing masked pixels with zero or median values of the surrounding pixels

    Returns
    -------
    random mask generator

    """
    batch_vol = (batch_size,) + image.shape[1:]
    img_output = deepcopy(image)

    j = 0
    num_cycle = np.ceil(img_output.shape[0] / batch_size)
    lprint(f'Masked pixels are replaced by {replace_by}')
    while True:
        i = np.mod(j, num_cycle).astype(int, copy=False)
        image_batch = img_output[batch_size * i : batch_size * (i + 1)]

        mask = masker(
            batch_vol, p=p_maskedpixels
        )  # generate a 2D mask with same size of image_batch; p of the pix are 0
        train_img = (
            mask * image_batch
        )  # pixels in training image are blocked by multiply by 0
        masknega = ~mask  # p of the pixels are 1

        if replace_by == 'random':
            train_img = train_img + np.random.random(batch_vol) * masknega
        elif replace_by == 'median':
            train_img = train_img + med_filter(image_batch) * masknega

        target_img = masknega * image_batch
        j += 1

        yield {'input': train_img, 'input_msk': masknega.astype(np.float32)}, target_img



import gc
import os
from abc import ABC, abstractmethod
from os.path import join

import jsonpickle
import numpy

from aydin.util.misc.json import encode_indent
from aydin.util.log.log import lprint


class RegressorBase(ABC):
    """Regressor base class"""

    def __init__(self):
        super().__init__()
        self._stop_fit = False
        self.num_channels = None
        self.models = 0

        self.loss_history = []

    def recommended_max_num_datapoints(self) -> int:
        """Recommended maximum number of datapoints

        Returns
        -------
        int

        """
        # very large number, essentially no limit by default
        return 1e9

    @abstractmethod
    def _fit(self, x_train, y_train, x_valid, y_valid, regressor_callback=None):
        """Fits function y=f(x) given training pairs (x_train, y_train).
        Stops when performance stops improving on the test dataset: (x_test, y_test).

        Parameters
        ----------
        x_train
            x training values
        y_train
            y training values
        x_valid
            x validation values
        y_valid
            y validation values
        regressor_callback
            regressor callback

        """

    def fit(
        self, x_train, y_train, x_valid=None, y_valid=None, regressor_callback=None
    ):
        """Fits function y=f(x) given training pairs (x_train, y_train).
        Stops when performance stops improving on the test dataset: (x_test, y_test).
        The target y_train can have multiple 'channels'. This will cause multiple regressors
        to be instanciated internally to be able to predict these channels from the input features.

        Parameters
        ----------
        x_train
            x training values
        y_train
            y training values
        x_valid
            x validation values
        y_valid
            y validation values
        regressor_callback
            regressor callback

        """
        has_more_than_one_channel = len(y_train.shape) > 1

        if x_valid is None:
            x_valid = x_train
            y_valid = y_train

        # to the multi-channel form, but with just one chanel;
        if not has_more_than_one_channel:
            y_train = y_train[numpy.newaxis, ...]
            y_valid = y_valid[numpy.newaxis, ...]

        self.models = []
        self._stop_fit = False
        for y_train_channel, y_valid_channel in zip(y_train, y_valid):

            gc.collect()
            model_channel = self._fit(
                x_train, y_train_channel, x_valid, y_valid_channel, regressor_callback
            )
            self.models.append(model_channel)
            self.loss_history.append(model_channel.loss_history)

        self.num_channels = len(self.models)

    def predict(self, x, models_to_use=None):
        """Predicts y given x by applying the learned function f: y=f(x)
        If the regressor is trained on multiple ouput channels, this will
        return the corresponding number of channels...

        Parameters
        ----------
        x
            x values
        models_to_use

        Returns
        -------
        numpy.typing.ArrayLike
            inferred y values

        """
        if models_to_use is None:
            models_to_use = self.models

        return numpy.stack([model.predict(x) for model in models_to_use])

    def stop_fit(self):
        """Stops training (can be called by another thread)"""
        self._stop_fit = True

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['models']
        return state

    def save(self, path: str):
        """Saves an 'all-batteries-included' regressor at a given path (folder).

        Parameters
        ----------
        path
            path to save to

        Returns
        -------
        frozen : Encoded JSON object

        """
        os.makedirs(path, exist_ok=True)

        frozen = encode_indent(self)

        lprint(f"Saving regressor to: {path}")
        with open(join(path, "regressor.json"), "w") as json_file:
            json_file.write(frozen)

        for i, model in enumerate(self.models):
            channel_path = join(path, f"channel{i}")
            os.makedirs(channel_path, exist_ok=True)
            frozen_model = encode_indent(model)
            with open(join(channel_path, "regressor_model.json"), "w") as json_file:
                json_file.write(frozen_model)

            model._save_internals(channel_path)

        return frozen

    @staticmethod
    def load(path: str):
        """Returns an 'all-batteries-included' regressor from a given path (folder).

        Parameters
        ----------
        path
            path to load from.

        Returns
        -------
        thawed : object

        """
        lprint(f"Loading regressor from: {path}")
        with open(join(path, "regressor.json"), "r") as json_file:
            frozen = json_file.read()

        thawed = jsonpickle.decode(frozen)

        thawed.models = []
        for i in range(thawed.num_channels):
            channel_path = join(path, f"channel{i}")
            lprint(f"Loading regressor model for channel {i} from: {path}")
            with open(join(channel_path, "regressor_model.json"), "r") as json_file:
                frozen_model = json_file.read()

            thawed_model = jsonpickle.decode(frozen_model)
            thawed_model._load_internals(channel_path)
            thawed.models.append(thawed_model)

        return thawed



"""
Axis codes:
       'X' width, 'Y' height, 'S' sample, 'I' image series|page|plane,
       'Z' depth, 'C' color|em-wavelength|channel, 'E' ex-wavelength|lambda,
       'T' time, 'R' region|tile, 'A' angle, 'P' phase, 'H' lifetime,
       'L' exposure, 'V' event, 'Q' unknown, '_' missing
"""

import os
import pathlib
import traceback
from contextlib import contextmanager
from os.path import exists
from pathlib import Path
import numpy
import skimage
import zarr
from czifile import czifile, CziFile
from nd2reader import ND2Reader
from numpy import array_equal
from tifffile import tifffile, TiffFile, memmap

from aydin.io.utils import is_zarr_storage, read_zarr_array
from aydin.util.log.log import lsection, lprint


def is_batch(code, shape, axes):
    """Method to check if given axis code belongs to a batch dimension.

    Parameters
    ----------
    code : str
    shape : tuple
    axes : str

    Returns
    -------
    bool

    """
    # special case:
    if len(shape) == 3 and 'X' in axes and 'Y' in axes and 'I' == code:
        return False

    return code not in 'XYZTQC'


def is_channel(code, length):
    """Method to check if given axis code belongs to channel dimension.

    Parameters
    ----------
    code :

    Returns
    -------
    bool

    """
    return code == "C" and not length > 8


class FileMetadata:
    """Metadata class used across aydin package.

    # TODO: make this docstring more detailed
    """

    def __init__(self):
        self.is_folder = None
        self.extension = None
        self.axes = None
        self.shape = None
        self.dtype = None
        self.format = None
        self.batch_axes = None
        self.channel_axes = None
        self.other = None
        self.splitted = False

    def __str__(self) -> str:
        return f" is_folder={self.is_folder}, ext={self.extension}, axes={self.axes}, shape={self.shape}, batch_axes={self.batch_axes}, channel_axes={self.channel_axes}, dtype={self.dtype}, format={self.format} "

    def __eq__(self, other):
        if not isinstance(other, FileMetadata):
            return NotImplemented  # don't attempt to compare against unrelated types

        return (
            self.is_folder == other.is_folder
            and self.extension == other.extension
            and self.axes == other.axes
            and self.shape == other.shape
            and self.dtype == other.dtype
            and self.format == other.format
            and self.batch_axes == other.batch_axes
            and self.channel_axes == other.channel_axes
            and self.other == other.other
        )


def imread(input_path):
    """Image reading method.

    Method takes the image path as a string argument. Upon certain
    checks and decisions it returns the image array and its
    corresponding metadata.

    Parameters
    ----------
    input_path : str

    Returns
    -------
    tuple(numpy.typing.ArrayLike, FileMetadata)
        Returns tuple of (array, metadata).

    """

    with lsection(f"Reading image file at: {input_path}"):

        metadata = FileMetadata()

        metadata.is_folder = os.path.isdir(input_path)
        metadata.extension = ((Path(input_path).suffix)[1:]).lower()

        is_tiff = 'tif' in metadata.extension or 'tiff' in metadata.extension
        is_czi = 'czi' in metadata.extension
        is_png = 'png' in metadata.extension
        is_jpg = 'jpg' in metadata.extension or 'jpeg' in metadata.extension
        is_zarr = 'zarr' in metadata.extension or is_zarr_storage(input_path)
        is_npy = 'npy' in metadata.extension
        is_npz = 'npz' in metadata.extension
        is_nd2 = 'nd2' in metadata.extension
        is_globlist = '*' in input_path

        try:
            if is_zarr:
                g = zarr.open(
                    input_path, mode='r'
                )  # TODO: fix this, it crashes with zarr arrays
                if isinstance(g, zarr.Array):
                    lprint(f"Reading file {input_path} as ZARR array")

                    if 'axes' in g.attrs:
                        metadata.axes = g.attrs['axes']
                else:
                    # Then we treat it as dexp-convention zarr group
                    lprint(f"Reading file {input_path} as ZARR group")
                    nb_arrays = 0
                    for key in g.group_keys():
                        nb_arrays += 1
                        if 'axes' in g[key][key].attrs:
                            metadata.axes = g.attrs['axes']

                metadata.format = 'zarr'
                array = read_zarr_array(input_path)
                metadata.shape = array.shape
                metadata.dtype = array.dtype

            elif is_tiff:
                lprint(f"Reading file {input_path} as TIFF file")
                with TiffFile(input_path) as tif:
                    if len(tif.series) >= 1:
                        serie = tif.series[0]
                        metadata.shape = serie.shape
                        metadata.dtype = serie.dtype
                        metadata.axes = serie.axes
                        metadata.other = tif.imagej_metadata
                    else:
                        lprint(f'There is no series in file: {input_path}')

                metadata.format = 'tiff'
                array = tifffile.imread(input_path)

            elif is_czi:
                lprint(f"Reading file {input_path} as CZI file")
                with CziFile(input_path) as czi:
                    metadata.format = 'czi'
                    metadata.axes = czi.axes
                    metadata.other = czi.metadata(raw=False)

                array = czifile.imread(input_path)
                metadata.shape = czi.shape
                metadata.dtype = czi.dtype

            elif is_png or is_jpg:
                lprint(f"Reading file {input_path} as PNG file")
                array = skimage.io.imread(input_path)

                # We check if this is a gray level image:
                if len(array.shape) == 3:
                    if array_equal(array[..., 0], array[..., 1]) and array_equal(
                        array[..., 0], array[..., 2]
                    ):
                        # We keep the first channel only:
                        array = array[..., 0]

                metadata.format = 'png' if is_png else 'jpg'
                metadata.shape = array.shape
                metadata.dtype = array.dtype

                if len(array.shape) == 2:
                    metadata.axes = "YX"
                elif len(array.shape) == 3 and array.shape[-1] in (3, 4):
                    metadata.axes = "YXC"
                elif len(array.shape) == 3:
                    metadata.axes = "ZYX"
                elif len(array.shape) == 4 and array.shape[-1] in (3, 4):
                    metadata.axes = "ZYXC"
                else:
                    metadata.axes = "ZYXC"
                    lprint(
                        f"Warning: Can't interpret {'png' if is_png else 'jpg'} structure, might be incorrect!"
                    )
            elif is_npy:
                lprint(f"Reading file {input_path} as NPY file")

                array = numpy.load(input_path)
                metadata.format = 'npy'
                metadata.shape = array.shape
                metadata.dtype = array.dtype
                metadata.axes = ''.join(('Q',) * len(array.shape))

            elif is_npz:
                lprint(f"Reading file {input_path} as NPZ file")

                data = numpy.load(input_path)
                lprint(data.files)

                # this could contain several arrays, we read the one with the most voxels (good heuristic):
                # We read the largest array:
                biggest_size = 0
                for _file in data.files:
                    _array = data[_file]
                    size = numpy.size(_array)
                    lprint(
                        f"Reading array of name: {_file}, shape: {_array.shape}, and dtype: {_array.dtype}, size: {size}"
                    )

                    if biggest_size < size:
                        lprint("Bigger!")
                        file = _file
                        biggest_size = size
                        array = _array

                # makse sure the array is 'clean':
                array = numpy.asarray(array)
                lprint(
                    f"Selected array: name: {file}, shape: {array.shape}, and dtype: {array.dtype}"
                )
                metadata.format = 'npz'
                metadata.shape = array.shape
                metadata.dtype = array.dtype
                metadata.axes = ('TZYX' + ''.join(('Q',) * len(array.shape)))[
                    0 : array.ndim
                ]

            elif is_nd2:
                lprint(f"Reading file {input_path} as ND2 file")
                import pims

                n2image = ND2Reader(input_path)

                metadata.format = 'nd2'
                metadata.axes = ''.join(n2image.axes).upper()  # TODO: check order!

                n2image.bundle_axes = n2image.axes
                array = numpy.asarray(n2image[0], dtype=metadata.dtype)
                metadata.shape = array.shape
                metadata.dtype = array.dtype

            elif is_globlist:
                lprint(f"Reading file {input_path} as file list")
                import pims

                array = pims.ImageSequence(input_path)
                metadata.format = 'globlist'
                metadata.shape = array.shape
                metadata.dtype = array.dtype
                metadata.axes = ('TZYX' + ''.join(('Q',) * len(array.shape)))[
                    0 : array.ndim
                ]

            # elif is_folder:
            #     from aydin.io import io
            #
            #     files = get_files_with_most_frequent_extension(input_path)
            #     files.sort()
            #
            #     imread = dask.delayed(io.imread, pure=True)  # Lazy version of imread
            #
            #     lazy_images = [
            #         imread(join(input_path, filename)) for filename in files
            #     ]  # Lazily evaluate imread on each path
            #
            #     file_metadata = analyse(join(input_path, files[0]))
            #
            #     arrays = [
            #         dask.array.from_delayed(
            #             lazy_image,  # Construct a small Dask array
            #             dtype=file_metadata.dtype,  # for every lazy value
            #             shape=file_metadata.shape,
            #         )
            #         for lazy_image in lazy_images
            #     ]
            #
            #     array = dask.array.stack(arrays, axis=0)
            #
            #     metadata.format = 'folder-' + file_metadata.format
            #     metadata.shape = array.shape
            #     metadata.dtype = array.dtype
            #     metadata.axes = 'Q' + file_metadata.axes
            #
            #     metadata.array = array
            #
            #     pass

            else:
                try:
                    array = skimage.io.imread(input_path)
                    metadata.format = pathlib.Path(input_path).suffix
                    metadata.shape = array.shape
                    metadata.dtype = array.dtype
                    metadata.axes = ('TZYX' + ''.join(('Q',) * len(array.shape)))[
                        0 : array.ndim
                    ]
                except Exception as error:
                    lprint(error)
                    lprint(traceback.format_exc())
                    lprint(
                        f"Tried to open file {input_path} with skimage io but failed to obtain image."
                    )
                    return None, None

        except Exception as error:
            lprint(error)
            lprint(traceback.format_exc())
            lprint(f"Could not read file {input_path} !")
            return None, None

        if metadata.axes:
            metadata.batch_axes = tuple(
                is_batch(axis, metadata.shape, metadata.axes) for axis in metadata.axes
            )

            metadata.channel_axes = tuple(
                is_channel(axis, s) for axis, s in zip(metadata.axes, metadata.shape)
            )

        lprint(f"Metadata: {metadata}")

        _sync_array_with_metadata(array, metadata)

        return array, metadata


def _sync_array_with_metadata(array, metadata):
    # We need to check if the metadata matches what we actually get, otherwise we need to update it.
    # This can happen for tiff files that are multi-part.
    if metadata is not None and array is not None:
        if metadata.shape != array.shape:
            metadata.shape = array.shape
        if metadata.dtype != array.dtype:
            metadata.dtype = array.dtype


def imwrite(array, output_path, metadata=None, overwrite=True):
    """Image writing method.

    Parameters
    ----------
    array : numpy.typing.ArrayLike
    output_path : str
    metadata : FileMetadata
    overwrite : bool

    """

    if not overwrite and exists(output_path):
        return

    if "png" in output_path and (
        len(array.shape) > 3
        or (len(array.shape) == 3 and array.shape[-1] not in [3, 4])
    ):
        lprint(
            "png images with more than 2 dimensions are not supported, will be writing the result as a tif"
        )
        output_path = f"{output_path[:output_path.rfind('.')]}.tif"

    if output_path[-3:] == "tif":
        _write_tiff(output_path, array, metadata)
    else:
        try:
            skimage.io.imsave(output_path, array)
        except Exception:
            # if skimage.io.imsave fails for any reason we fallback to .tif format
            output_path = f"{output_path[:output_path.rfind('.')]}.tif"
            _write_tiff(output_path, array, metadata)


def _write_tiff(output_path, array, metadata):
    """Internal method to write .tiff files with given array and metadata.

    Parameters
    ----------
    array : numpy.typing.ArrayLike
    output_path : str
    metadata : FileMetadata

    """
    # We get the ij metadata:
    ijmetadata = None if metadata is None else metadata.other

    # Normalise to {}:
    ijmetadata = {} if ijmetadata is None else ijmetadata

    tifffile.imwrite(output_path, array, metadata=ijmetadata)


@contextmanager
def mapped_tiff(output_path, shape, dtype):
    """Mapped tiff context manager.

    Parameters
    ----------
    output_path
    shape
    dtype

    Yields
    ------

    """
    array = memmap(output_path, shape=shape, dtype=dtype)
    try:
        yield array
        array.flush()
    finally:
        del array
        lprint(
            f"Flushing and writing all bytes to TIFF file {output_path}  (shape={shape}, dtype={dtype})"
        )



import numpy as np
import tensorflow as tf
from tensorflow.python.keras import backend
from tensorflow.python.keras.layers import Lambda


def Split(
    x, idx, batchsize=1, lyrname=None
):  # TODO: refactor into a class or a util function
    """
    Split tensor at the batch axis. Only for shift convolution architecture.

    Parameters
    ----------
    x
        input tensor
    idx
        index for the split chunk
    batchsize
        batch size
    lyrname : string
        layer name

    Returns
    -------
    Split layer : layers.Lambda

    """
    out_shape = backend.int_shape(x[0])
    return Lambda(
        lambda xx: xx[idx * batchsize : (idx + 1) * batchsize],
        output_shape=out_shape,
        name=lyrname,
    )


def Swish(name=None):
    """
    Swish Layer

    Parameters
    ----------
    name : string

    Returns
    -------
    Swish layer : layers.Lambda

    """
    return Lambda(tf.nn.swish, name=name)


def Rot90(xx, kk=1, lyrname=None):
    """
    Rotate tensor by 90 degrees for 2D, 3D images. Only for shift convolution architecture.

    Parameters
    ----------
    xx
        input tensor from previous layer
    kk
        index for rotation (crock wise)
    lyrname : string
        name of the layer

    Returns
    -------
    Rot90 layer : layers.Lambda
    """
    out_shape = list(backend.int_shape(xx))
    if kk % 2 == 1 and 0 < kk % 6 < 5:
        out_shape[-3:-1] = np.flip(out_shape[-3:-1], 0)
    elif abs(kk) % 6 == 5 or kk % 6 == 0:
        out_shape[1:3] = np.flip(out_shape[1:3], 0)
    if len(out_shape) == 4:
        tp_axis = [0, 2, 1, 3]  # (batch, longitudinal, horizontal, channel)
    elif len(out_shape) == 5:
        tp_axis = [
            0,
            1,
            3,
            2,
            4,
        ]  # (batch, z-direction, longitudinal, horizontal, channel)
        tp_axis2 = (0, 3, 2, 1, 4)  # rotation along another axis
    else:
        raise ValueError(
            'Input shape has to be 4D or 5D. e.g. (Batch, (depth), longitudinal, horizontal, channel)'
        )

    if kk < 0:
        direction = [-2, -3, -2, 1]
    else:
        direction = [-3, -2, 1, -2]
    if abs(kk) % 6 == 5 and len(out_shape) == 5:
        return Lambda(
            lambda xx: backend.reverse(
                backend.permute_dimensions(xx, tp_axis2), axes=direction[2]
            ),
            output_shape=out_shape[1:],
            name=lyrname,
        )
    elif kk % 6 == 0 and len(out_shape) == 5 and kk != 0:
        return Lambda(
            lambda xx: backend.reverse(
                backend.permute_dimensions(xx, tp_axis2), axes=direction[3]
            ),
            output_shape=out_shape[1:],
            name=lyrname,
        )
    else:
        if kk % 4 == 1:
            return Lambda(
                lambda xx: backend.reverse(
                    backend.permute_dimensions(xx, tp_axis), axes=direction[0]
                ),
                output_shape=out_shape[1:],
                name=lyrname,
            )
        elif kk % 4 == 2:
            return Lambda(
                lambda xx: backend.reverse(backend.reverse(xx, axes=-2), axes=-3),
                output_shape=out_shape[1:],
                name=lyrname,
            )
        elif kk % 4 == 3:
            return Lambda(
                lambda xx: backend.reverse(
                    backend.permute_dimensions(xx, tp_axis), axes=direction[1]
                ),
                output_shape=out_shape[1:],
                name=lyrname,
            )
        elif kk % 4 == 0:
            return Lambda(lambda xx: xx, output_shape=out_shape[1:], name=lyrname)



import json
import jsonpickle


def encode_indent(object):
    """
    Method to encode JSON

    Parameters
    ----------
    object : object

    Returns
    -------
    JSONEncoder
        Encoded JSON object

    """
    return json.dumps(json.loads(jsonpickle.encode(object)), indent=4, sort_keys=True)


def save_any_json(dict2save, path):
    frozen = jsonpickle.encode(dict2save)
    with open(path, 'w') as fp:
        json.dump(frozen, fp)


def load_any_json(path):
    with open(path, 'r') as fp:
        read_dict = json.load(fp)

    return jsonpickle.decode(read_dict)



from aydin.it.deconvolution.lr_deconv_scipy import ImageTranslatorLRDeconvScipy
from aydin.util.log.log import lsection
from aydin.util.misc.progress_bar import ProgressBar


class LucyRichardson:
    """LucyRichardson Service.

    Parameters
    ----------
    backend : str
        Opacity of the layer visual, between 0.0 and 1.0.
    psf_kernel : array_like
        2D or 3D kernel, dimensions should be odd numbers and numbers sum to 1
    max_num_iterations : int
    backend : str
        Computation backend selection.
    """

    def __init__(
        self, *, psf_kernel=None, max_num_iterations=20, backend='scipy', **kwargs
    ):

        # default:
        deconvolution_class = ImageTranslatorLRDeconvScipy
        if backend == 'scipy':
            deconvolution_class = globals()['ImageTranslatorLRDeconvScipy']
        elif backend == 'cupy':
            deconvolution_class = globals()['ImageTranslatorLRDeconvCupy']
        elif backend == 'gputools':
            deconvolution_class = globals()['ImageTranslatorLRDeconvGputools']

        self.it = deconvolution_class(
            psf_kernel=psf_kernel, max_num_iterations=max_num_iterations, **kwargs
        )

    def train(self, input_image, target_image, **kwargs):
        """Method to prepare ImageTranslatorLRDeconv instance.
        For now there is no actual training going on in lower level calls.
        However, calling this method needed to keep normal it compute flow.

        Parameters
        ----------
        input_image : numpy.ndarray
        target_image : numpy.ndarray

        """
        with lsection("LucyRichardson training is starting..."):
            self.it.train(
                input_image,
                target_image,
                train_valid_ratio=kwargs['train_valid_ratio']
                if 'train_valid_ratio' in kwargs
                else 0.1,
                callback_period=kwargs['callback_period']
                if 'callback_period' in kwargs
                else 3,
                jinv=kwargs['jinv'] if 'jinv' in kwargs else None,
            )

    def deconvolve(self, input_image, **kwargs):
        """Method to deconvolve an image by LucyRichardson technique.

        Parameters
        ----------
        input_image : numpy.ndarray

        Returns
        -------
        response : numpy.ndarray
            deconvolved image.

        """
        with lsection("LucyRichardson translation is starting..."):
            response = self.it.translate(
                input_image,
                tile_size=kwargs['tile_size'] if 'tile_size' in kwargs else None,
            )
            response = response.astype(input_image.dtype, copy=False)

            return response


def lucyrichardson(input_image, *, batch_axes=None, chan_axes=None, backend=None):
    """Method to denoise an image with trained Noise2Self.

    Parameters
    ----------
    input_image : numpy.ndarray
        Input image to deconvolve.
    batch_axes : array_like, optional
        Indices of batch axes.
    chan_axes : array_like, optional
        Indices of channel axes.

    Returns
    -------
    response : numpy.ndarray

    """
    pbar = ProgressBar(total=100)
    lr = LucyRichardson(pbar, backend=backend)

    # Train
    lr.train(input_image)

    # Denoise
    deconvolved = lr.deconvolve(input_image)

    return deconvolved



from typing import Tuple, List

import numpy
import scipy
from numpy import argmax, unravel_index
from numpy.linalg import norm
from scipy.ndimage import gaussian_filter

from aydin.util.crop.rep_crop import representative_crop


def auto_detect_blindspots(
    image,
    batch_axes: Tuple[bool] = None,
    channel_axes: Tuple[bool] = None,
    threshold=0.01,
    max_blind_spots=3,
    max_range: int = 3,
    window: int = 31,
    crop_border: int = 2,
    fastmode: bool = True,
) -> Tuple[List[Tuple[int, ...]], numpy.ndarray]:
    """Automatically determines the list of blind-spots for Noise2Self

    Parameters
    ----------
    image : numpy.typing.ArrayLike
        Image for which to conduct blind-spot analysis

    batch_axes : Tuple[bool]
        Batch axes for image

    channel_axes : Tuple[bool]
        Channel axes for image

    threshold
        threshold for inclusion as a blind-spot

    max_blind_spots
        Max number of blindspots

    max_range : int
        maximal range of the returned blind-spots and noise autocorrelogram

    window : int
        window used for computing autocorrelogram (best left unchanged)

    crop_border : int
        Cropping border

    fastmode : str
        When True we use an even faster approach.


    Returns
    -------
    Tuple of list of blindspots and the noise autocorrelogram

    """

    # Handle default values for batch and channel dim specification:
    if batch_axes is None:
        batch_axes = (False,) * image.ndim
    if channel_axes is None:
        channel_axes = (False,) * image.ndim

    # Ensure there is at least one batch or channel dimensions:
    if batch_axes == (False,) * image.ndim and channel_axes == (False,) * image.ndim:
        image = image[numpy.newaxis, ...]
        batch_axes = (True,) + batch_axes
        channel_axes = (False,) + channel_axes

    # we collapse all batch and channel dimensions together:
    shape = tuple(
        s for s, b, c in zip(image.shape, batch_axes, channel_axes) if not b and not c
    )
    if len(shape) < image.ndim:
        shape = (-1,) + shape
    image = numpy.reshape(image, shape)

    # And pick the image with the most variance:
    chosen_image = None
    chosen_image_variance = -1
    for image_index in range(image.shape[0]):
        one_image = image[image_index]
        variance = numpy.var(one_image)
        if variance > chosen_image_variance and not numpy.isnan(variance):
            chosen_image_variance = variance
            chosen_image = one_image

    # chosen_image should not be None, but if does happen (happened once!) then let's play safe:
    if chosen_image is not None:
        image = chosen_image

    # First we need to remove the borders of the image, as sometimes the borders have artefacts:
    if crop_border > 0 and all(s > 2 * crop_border for s in image.shape):
        crop_slice = tuple(
            slice(max(s // 16, crop_border), -max(s // 16, crop_border))
            for s in image.shape
        )
        image = image[crop_slice]

    if fastmode:
        # obtain representative crop, to speed things up...
        image = representative_crop(image, crop_size=int(1e6), favour_odd_lengths=True)

    # We compute the autocorrelogram of the noise:
    noise_auto = noise_autocorrelation(image, max_range=max_range, window=window)

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(noise_auto, name='noise_auto')

    # What is the intensity of the nth strongest correlation?
    noise_auto_flat = noise_auto.flatten()
    noise_auto_flat.sort()
    nth_strongest_correlation = noise_auto_flat[-max_blind_spots]

    # We adjust the threshold to take into account the max number of blindspots requested:
    threshold = max(threshold, nth_strongest_correlation)

    # We list the blind spots:
    blind_spots = []
    for idx, x in numpy.ndenumerate(noise_auto):
        if x >= threshold:
            blind_spot = tuple(n - max_range for n in idx)
            blind_spots.append(blind_spot)

    # Remove any constant offset per dimension:
    blind_spots = numpy.array(blind_spots)
    blind_spots = blind_spots - numpy.mean(blind_spots, axis=0, keepdims=True).astype(
        dtype=numpy.int32
    )

    # Convert back to list of tuples:
    blind_spots = list(tuple(a) for a in blind_spots)

    return blind_spots, noise_auto


def noise_autocorrelation(image, max_range: int = 3, window: int = 31) -> numpy.ndarray:
    """This function computes the noise autocorrelogram.

    Principle: We simply divide the autocorrelogram of the raw image by the autocorrelogram of a naively denoised image.
    What is left is the autocorrelogram of the noise.

    Parameters
    ----------
    image : numpy.typing.ArrayLike
        image to compute the noise autocorrelation
    max_range : int
        expected maximal range of correlation (3 is a good number!)
    window : int
        window size for correlation computation (31 is a good number!)

    Returns
    -------
    analysis : numpy.ndarray
        noise autocorrelogram of shape (max_range*2+1,)*ndim  where ndim is the number of dimensions of the input image.
    """

    # Enfor

    # First we compute the auto-correlation of the raw image:
    auto_corr = _autocorrelation(image, window=window)

    # Second we compute the auto-correlation of the image after 'rough' denoising:
    blurred_image = gaussian_filter(image, sigma=0.5)
    # blurred_image = median_filter(blurred_image, size=2)
    blurred_auto_corr = _autocorrelation(blurred_image, window=window)

    # We 'remove' by division the autocorrelogram fixed_pattern that is common to both:
    analysis = auto_corr / blurred_auto_corr

    # Now we compute the maximum intensity outside of the central region:
    outside = analysis.copy()
    # center = tuple((min(s, window) - 1) // 2 for s in analysis.shape)
    center = unravel_index(argmax(analysis), analysis.shape)

    # Max range might be too much for very shallow dimensions, and for non odd dimensions the center
    # is not at the cente so we need to reduce the range accordingly:
    range = tuple(min(c, s - c, max_range) for c, s in zip(center, analysis.shape))

    # Let's compute the center slice:
    center_slice = tuple(slice(c - r, c + r + 1, 1) for c, r in zip(center, range))
    outside[center_slice] = 0
    floor = numpy.max(outside)

    # And remove it:
    analysis = analysis - floor
    analysis = analysis.clip(0, numpy.math.inf)

    # We normalise to a sum of 1:
    analysis /= analysis.max()

    # Now we have the noise autocorelogram, we can crop that to the expected max range:
    analysis = analysis[center_slice]

    return analysis


def _autocorrelation(image, window: int = 31) -> numpy.ndarray:
    """Computes the autocorrelation of an image over a cropped window around the origin

    Parameters
    ----------
    image : numpy.typing.ArrayLike
        image
    window : int
        window size

    Returns
    -------
    array : numpy.ndarray

    """
    image = image.astype(numpy.float32)
    image /= norm(image)

    array = _phase_correlation(image, image)
    shift = tuple(min(window, s) // 2 for s in image.shape)
    array = numpy.roll(array, shift=shift, axis=range(image.ndim))
    slice_tuple = (slice(0, window),) * image.ndim
    array = array[slice_tuple]
    return array


def _phase_correlation(image, reference_image) -> numpy.ndarray:
    """Computes the phase correlation between am image and a reference image.
    Note: both images must be of the same shape.

    Parameters
    ----------
    image : numpy.typing.ArrayLike
    reference_image : numpy.ArrayLike

    Returns
    -------
    r : numpy.ndarray

    """
    G_a = scipy.fft.fftn(image, workers=-1)
    G_b = scipy.fft.fftn(reference_image, workers=-1)
    conj_b = numpy.ma.conjugate(G_b)
    R = G_a * conj_b
    r = numpy.absolute(scipy.fft.ifftn(R, workers=-1))
    return r



import random
from math import pi, cos, sin

import matplotlib.pyplot as plt
import numpy as np
from skimage import data
from skimage.color import rgb2gray
from skimage.draw import circle, circle_perimeter, line_aa

from aydin.analysis.correlation import correlation, correlation_distance


def discs(shape=(512, 512), radius=10, num=512, fill=True, image=None):
    if image is None:
        image = np.zeros(shape, dtype=np.float)

    for i in range(num):
        x = int(random.uniform(0, shape[0]))
        y = int(random.uniform(0, shape[1]))
        if fill:
            rr, cc = circle(x, y, radius, shape=image.shape)
            image[rr, cc] = 1
        else:
            rr, cc = circle_perimeter(
                x, y, radius, method='bresenham', shape=image.shape
            )
            image[rr, cc] = 1

    return image


def lines(shape=(512, 512), length=10, num=512, fill=False, image=None):
    if image is None:
        image = np.zeros(shape, dtype=np.float)

    for i in range(num):
        xb = int(random.uniform(length, shape[0] - 1 - length))
        yb = int(random.uniform(length, shape[1] - 1 - length))
        angle = random.uniform(0, 2 * pi)
        xe = int(xb + length * cos(angle))
        ye = int(yb + length * sin(angle))
        rr, cc, val = line_aa(xb, yb, xe, ye)
        image[rr, cc] = val

    return image


plot_length = 256


def noise_example():
    image = np.random.random((1500, 1500))
    plt.imshow(image)
    plt.show()

    plt.plot(correlation(image)[0][:32], label='y')
    plt.plot(correlation(image)[1][:32], label='x')
    plt.legend()
    plt.show()

    print("noise  %d,%d " % correlation_distance(image))


def camera_example():
    camera = data.camera()
    plt.imshow(camera)
    plt.show()

    plt.plot(correlation(camera)[0][:plot_length], label='y')
    plt.plot(correlation(camera)[1][:plot_length], label='x')
    plt.legend()
    plt.show()

    print("camera  %d,%d " % correlation_distance(camera))


def astronaut_example():
    astronaut = rgb2gray(data.astronaut())
    plt.imshow(astronaut)
    plt.show()

    plt.plot(correlation(astronaut)[0][:plot_length], label='y')
    plt.plot(correlation(astronaut)[1][:plot_length], label='x')
    plt.legend()
    plt.show()

    print("astronaut  %d,%d " % correlation_distance(astronaut))


def clock_example():
    clock = data.clock()[70:220, 150:270]
    plt.imshow(clock)
    plt.show()

    plt.plot(correlation(clock)[0][:plot_length], label='y')
    plt.plot(correlation(clock)[1][:plot_length], label='x')
    plt.legend()
    plt.show()

    print("clock  %d,%d " % correlation_distance(clock))


def coins_example():
    coins = data.coins()
    plt.imshow(coins)
    plt.show()

    plt.plot(correlation(coins)[0][:plot_length], label='y')
    plt.plot(correlation(coins)[1][:plot_length], label='x')
    plt.legend()
    plt.show()

    print("coins  %d,%d " % correlation_distance(coins))


def discs_example():
    discs1 = discs(shape=(1500, 1500), radius=40, num=90)
    discs2 = discs(shape=(1500, 1500), radius=20, num=160)
    discs3 = discs(shape=(1500, 1500), radius=10, num=420)

    plt.imshow(discs1)
    plt.show()
    plt.imshow(discs2)
    plt.show()
    plt.imshow(discs3)
    plt.show()

    plt.plot(correlation(discs1)[0][:plot_length], label='y 40')
    plt.plot(correlation(discs2)[0][:plot_length], label='y 20')
    plt.plot(correlation(discs3)[0][:plot_length], label='y 10')
    plt.plot(correlation(discs1)[1][:plot_length], label='x 40')
    plt.plot(correlation(discs2)[1][:plot_length], label='x 20')
    plt.plot(correlation(discs3)[1][:plot_length], label='x 10')
    plt.legend()
    plt.show()

    print("discs1  %d,%d " % correlation_distance(discs1))
    print("discs2  %d,%d " % correlation_distance(discs2))
    print("discs3  %d,%d " % correlation_distance(discs3))


def lines_example():
    lines1 = lines(shape=(1500, 1500), length=40, num=3000)
    lines2 = lines(shape=(1500, 1500), length=20, num=6000)
    lines3 = lines(shape=(1500, 1500), length=10, num=12000)

    plt.imshow(lines1)
    plt.show()
    plt.imshow(lines2)
    plt.show()
    plt.imshow(lines3)
    plt.show()

    plt.plot(correlation(lines1)[0][:plot_length], label='40')
    plt.plot(correlation(lines2)[0][:plot_length], label='20')
    plt.plot(correlation(lines3)[0][:plot_length], label='10')
    plt.legend()
    plt.show()

    print("lines1  %d,%d " % correlation_distance(lines1))
    print("lines2  %d,%d " % correlation_distance(lines2))
    print("lines3  %d,%d " % correlation_distance(lines3))


noise_example()
camera_example()
astronaut_example()
clock_example()
coins_example()
discs_example()
lines_example()



import math
from typing import Optional

import numpy
from sklearn.decomposition import SparseCoder

from aydin.util.crop.rep_crop import representative_crop
from aydin.util.dictionary.dictionary import (
    fixed_dictionary,
    extract_normalised_vectorised_patches,
)
from aydin.util.j_invariance.j_invariant_classic import calibrate_denoiser_classic
from aydin.util.j_invariance.j_invariant_smart import calibrate_denoiser_smart
from aydin.util.log.log import lsection, lprint
from aydin.util.patch_size.patch_size import default_patch_size
from aydin.util.patch_transform.patch_transform import reconstruct_from_nd_patches


def calibrate_denoise_dictionary_fixed(
    image,
    patch_size: int = None,
    try_omp: bool = True,
    try_lasso_lars: bool = False,
    try_lasso_cd: bool = False,
    try_lars: bool = False,
    try_threshold: bool = False,
    num_sparsity_values_to_try: int = 6,
    dictionaries: str = 'dct',
    crop_size_in_voxels: Optional[int] = None,
    max_num_evaluations: int = 256,
    display_dictionary: bool = False,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the dictionary-based denoiser for the given image and returns the
    optimal parameters obtained using the N2S loss.

    Parameters
    ----------
    image : ArrayLike
        Image to calibrate denoiser for.

    patch_size : int
        Patch size. Common parameter to both 'learned',
        or 'fixed' dictionary types.
        (advanced)

    try_omp: bool
        Whether OMP should be tried as a sparse coding
        algorithm during calibration.

    try_lasso_lars: bool
        Whether LASSO-LARS should be tried as a sparse
        coding algorithm during calibration.

    try_lasso_cd: bool
        Whether LASSO-CD should be tried as a sparse
        coding algorithm during calibration.

    try_lars: bool
        Whether LARS should be tried as a sparse coding
        algorithm during calibration.

    try_threshold: bool
        Whether 'threshold'' should be tried as a sparse
        coding algorithm during calibration.

    num_sparsity_values_to_try: int
        Maximum number of sparsity values to try during calibration
        (advanced)

    dictionaries: str
        Fixed dictionaries to be included. Can be: 'dct',
        'dst'.

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate
        denoiser.
        (advanced)

    max_num_evaluations: int
        Maximum number of evaluations for finding the
        optimal parameters.
        (advanced)

    display_dictionary: bool
        If True displays dictionary with napari -- for
        debug purposes.

    display_images: bool
        When True the denoised images encountered during
        optimisation are shown.

    other_fixed_parameters: dict
        Any other fixed parameters


    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """
    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # Normalise patch size:
    patch_size = default_patch_size(image, patch_size, odd=True)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels)

    # Partial function:
    def _denoise_dictionary(
        image, max_freq: float = 0.5, coding_mode: str = 'omp', **parameters
    ):
        dictionary = fixed_dictionary(
            image, patch_size=patch_size, dictionaries=dictionaries, max_freq=max_freq
        )
        denoised_image = denoise_dictionary_fixed(
            image, dictionary=dictionary, coding_mode=coding_mode, **parameters
        )
        return denoised_image

    # coding modes to try:
    coding_modes = []
    if try_omp:
        coding_modes.append('omp')
    if try_lasso_lars:
        coding_modes.append('lasso_lars')
    if try_lasso_cd:
        coding_modes.append('lasso_cd')
    if try_lars:
        coding_modes.append('lars')
    if try_threshold:
        coding_modes.append('threshold')

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {
        'max_freq': (0.01, 1.3),
        # numpy.arange(0.05, 1.4, 0.05),  # (0.01, 1.3),
        'coding_mode': coding_modes,
        # 'lasso_lars', 'lasso_cd', 'lars', 'omp', 'threshold'
    }

    # Calibrate denoiser:
    best_parameters = calibrate_denoiser_smart(
        crop,
        _denoise_dictionary,
        denoise_parameters=parameter_ranges,
        max_num_evaluations=max_num_evaluations,
    )
    lprint(f"Best parameters: {best_parameters}")

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {
        'sparsity': [1, 2, 3, 4, 8, 16][:num_sparsity_values_to_try],
        # 'lasso_lars', 'lasso_cd', 'lars', 'omp', 'threshold'
    }

    # Calibrate denoiser:
    best_parameters = (
        calibrate_denoiser_classic(
            crop,
            _denoise_dictionary,
            denoise_parameters=parameter_ranges,
            other_fixed_parameters=best_parameters | other_fixed_parameters,
            display_images=display_images,
        )
        | best_parameters
        | other_fixed_parameters
    )

    # Cleaning up a bit:
    best_parameters.pop('other_fixed_parameters')

    lprint(f"Final best parameters: {best_parameters}")

    # we need to replace the max freq argument with the actual dictionary
    # because that's what our client facing denoise function expects:
    max_freq = best_parameters.pop('max_freq')

    # Dictionary to use based on fixed and best parameters:
    dictionary = fixed_dictionary(
        image, patch_size=patch_size, dictionaries=dictionaries, max_freq=max_freq
    )

    best_parameters = best_parameters | {'dictionary': dictionary}

    if display_dictionary:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(
                dictionary.reshape(len(dictionary), *patch_size), name='dictionary'
            )

    # Memory needed:
    memory_needed = 2 * image.nbytes + 6 * image.nbytes * math.prod(patch_size)

    return denoise_dictionary_fixed, best_parameters, memory_needed


def denoise_dictionary_fixed(
    image,
    dictionary=None,
    coding_mode: str = 'omp',
    sparsity: int = 1,
    gamma: float = 0.001,
    multi_core: bool = True,
    **kwargs,
):
    """
    Denoises the given image using sparse-coding over a fixed
    dictionary of nD image patches. The dictionary learning and
    patch sparse coding uses scikit-learn's Batch-OMP implementation.

    Parameters
    ----------
    image: ArrayLike
        nD image to be denoised

    dictionary: ArrayLike
        Dictionary to use for denosing image via sparse coding.
        By default (None) a fixed dictionary is used.

    coding_mode: str
        Type of sparse coding, can be:  'lasso_lars', 'lasso_cd', 'lars', 'omp',
        or 'threshold'

    sparsity: int
        How many atoms are used to represent each patch after denoising.

    gamma: float
        How much the periphery of teh patches contributes to the final denoised
        image. Larger gamma means that we keep more of the central pixels of the
        patches, smaller values lead to a more uniform contribution.
        A value of 1 corresponds to the default blackman window.

    multi_core: bool
        By default we use as many cores as possible, in some cases, for small
        (test) images, it might be faster to run on a single core instead of
        starting the whole parallelization machinery.

    Returns
    -------
    Denoised image
    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    if dictionary is None:
        # learn dictionary with all defaults:
        dictionary = fixed_dictionary(image)

    # we can infer patch shape from dictionary:
    patch_size = dictionary.shape[1:]

    with lsection(f"Denoise image of shape {image.shape} and dtype {image.dtype}"):
        # vectorise dictionary:
        vectorised_dictionary = dictionary.reshape(len(dictionary), -1)

        # setup sparse coder:
        coder = SparseCoder(
            vectorised_dictionary,
            transform_algorithm=coding_mode,
            transform_n_nonzero_coefs=sparsity,
            n_jobs=-1 if multi_core else 1,
        )

        # First we extract _all_ patches from the image, without any normalisation:
        with lsection("Extract all patches from image..."):
            patches, patch_means, _ = extract_normalised_vectorised_patches(
                image,
                patch_size=patch_size,
                max_patches=None,
                normalise_means=True,
                normalise_stds=False,
                output_norm_values=True,
            )

        with lsection("Obtain sparse codes for each patch..."):
            code = coder.transform(patches)

        with lsection("Reconstruct patches from codes..."):
            denoised_patches = numpy.dot(code, vectorised_dictionary)
            # Add back means:
            denoised_patches += patch_means

        with lsection("Reshape to patches..."):
            denoised_patches = denoised_patches.reshape(len(patches), *patch_size)

        with lsection("Reconstructing image from patches..."):
            # Reconstructs image from denoised patches:
            denoised_image = reconstruct_from_nd_patches(
                patches=denoised_patches, image_shape=image.shape, gamma=gamma
            )

    return denoised_image



import importlib
import inspect
import os
import pkgutil
import shutil
from abc import abstractmethod, ABC
from pathlib import Path

from aydin.util.log.log import lprint


class DenoiseRestorationBase(ABC):
    def __init__(self):
        pass

    @property
    @abstractmethod
    def configurable_arguments(self):
        """Returns the configurable arguments that will be exposed
        on GUI and CLI.
        """
        NotImplementedError()

    @property
    @abstractmethod
    def implementations(self):
        """Returns the list of discovered implementations for given method."""
        NotImplementedError()

    @abstractmethod
    def stop_running(self):
        """Method to stop running restoration instance."""
        NotImplementedError()

    @abstractmethod
    def train(
        self, noisy_image, *, batch_axes=None, chan_axes=None, image_path=None, **kwargs
    ):
        """Method to run training.

        Parameters
        ----------
        noisy_image : numpy.ndarray
        batch_axes : array_like, optional
            Indices of batch axes.
        chan_axes : array_like, optional
            Indices of channel axes.
        image_path : str

        Returns
        -------
        response : numpy.ndarray

        """
        NotImplementedError()

    @abstractmethod
    def denoise(self, noisy_image, *, batch_axes=None, chan_axes=None, **kwargs):
        """Method to denoise an image.

        Parameters
        ----------
        batch_axes : array_like, optional
            Indices of batch axes.
        chan_axes : array_like, optional
            Indices of channel axes.
        noisy_image : numpy.ndarray

        Returns
        -------
        response : numpy.ndarray

        """
        NotImplementedError()

    @staticmethod
    def get_implementations_in_a_module(module):
        return [
            x
            for x in pkgutil.iter_modules(tuple(module.__path__))
            if not x.ispkg and x.name != 'base'
        ]

    @staticmethod
    def get_class_implementation_kwonlyargs(package, module, implementation_class_name):
        response = importlib.import_module(package.__name__ + '.' + module.name)
        elem = [
            x for x in dir(response) if implementation_class_name.lower() in x.lower()
        ][
            0
        ]  # class name

        class_itself = response.__getattribute__(elem)
        fullargspec = inspect.getfullargspec(class_itself.__init__)

        args_offset = len(fullargspec.args) - len(fullargspec.defaults)

        args = {
            "arguments": fullargspec.args[args_offset:],
            "defaults": fullargspec.defaults,
            "annotations": fullargspec.annotations,
            "reference_class": class_itself,
        }

        return args

    @staticmethod
    def get_function_implementation_kwonlyargs(
        package, module, implementation_function_name
    ):
        response = importlib.import_module(package.__name__ + '.' + module.name)

        function_itself = response.__getattribute__(implementation_function_name)

        fullargspec = inspect.getfullargspec(function_itself)

        args_offset = len(fullargspec.args) - len(fullargspec.defaults)

        args = {
            "arguments": fullargspec.args[args_offset:],
            "defaults": fullargspec.defaults,
            "annotations": fullargspec.annotations,
            "reference_class": function_itself,
        }

        return args

    @staticmethod
    def clean_model_folder(model_folder):
        """Method to clean model folder created"""
        shutil.rmtree(model_folder)

    @staticmethod
    def archive_model(source, destination):
        """Archives the model to given destination.

        Parameters
        ----------
        source : str
        destination : str

        """
        name = Path(source).name
        format = "zip"
        archive_from = os.path.dirname(source)
        archive_to = os.path.basename(source.strip(os.sep))
        shutil.make_archive(name, format, archive_from, archive_to)
        if os.path.exists(os.path.join(destination, f"{name}.{format}")):
            lprint(
                "Previously existing model will be deleted before saving the new model"
            )
            os.remove(os.path.join(destination, f"{name}.{format}"))
        shutil.move(f"{name}.{format}", destination)

    def save_model(self, model_path):
        """Saves the latest trained model next to the input image file.

        Parameters
        ----------
        model_path : str

        """
        # Save the model first
        self.it.save(model_path)

        # Make archive for the model
        self.archive_model(model_path, os.path.dirname(model_path))

        # clean the model folder
        self.clean_model_folder(model_path)



import numpy

from aydin.features.standard_features import StandardFeatureGenerator


def demo_collect_feature_2d():
    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_spatial_features=True,
        include_median_features=True,
        include_dct_features=True,
        num_sinusoidal_features=4,
        include_random_conv_features=True,
    )

    image = numpy.zeros(shape=(17, 17), dtype=numpy.float32)
    image[8, 8] = 1
    image[::2, ::2] += 0.05

    # feature generator requires images in 'standard' form: BCTZYX
    norm_image = image[numpy.newaxis, numpy.newaxis, ...]

    features = generator.compute(
        norm_image,
        exclude_center_value=True,
        excluded_voxels=[(0, -1), (0, 0)],
        spatial_feature_offset=(0, 0),
        spatial_feature_scale=(1 / 17, 1 / 17),
    )

    assert features is not None

    features = features.squeeze()

    features = numpy.transpose(features, axes=(2, 0, 1))

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image', colormap='plasma')
        viewer.add_image(features, name='features', colormap='plasma')


demo_collect_feature_2d()



import numpy as np
import tensorflow as tf
from scipy.signal.signaltools import (
    _inputs_swap_needed,
    _init_freq_conv_axes,
    _apply_conv_mode,
)

from aydin.it.deconvolution.lr_deconv import ImageTranslatorLRDeconv


class ImageTranslatorLRDeconvTensorflow(ImageTranslatorLRDeconv):
    """Lucy Richardson Deconvolution, TensorFlow backend"""

    def __init__(self, *args, **kwargs):
        """Constructs a Lucy Richardson deconvolution image translator.

        Parameters
        ----------
        psf_kernel : numpy.typing.ArrayLike
            2D or 3D kernel, dimensions should be odd numbers and numbers sum to 1
        max_num_iterations : int
        clip : bool
        backend : str
            Computation backend selection.
        kwargs : dict
        """
        super().__init__(*args, **kwargs)

        self.max_voxels_per_tile = (
            320 ** 3 if self.max_voxels_per_tile is None else self.max_voxels_per_tile
        )

    def _convert_array_format_in(self, input_image):
        return input_image

    def _convert_array_format_out(self, output_image):
        return output_image

    def _get_convolution_method(self, input_image, psf_kernel):
        """Method to get convolution method.

        Parameters
        ----------
        input_image : array_like
        psf_kernel : array_like

        Returns
        -------
        convolve

        """
        return self._convolution

    def _get_pad_method(self):
        """Method to get pad method.

        Returns
        -------
        pad

        """
        import tensorflow

        return tensorflow.pad

    def _convolution(self, in1, in2, mode="full"):
        volume = np.asarray(in1)
        kernel = np.asarray(in2)

        if volume.ndim == kernel.ndim == 0:
            return volume * kernel
        elif volume.ndim != kernel.ndim:
            raise ValueError("volume and kernel should have the same dimensionality")

        if _inputs_swap_needed(mode, volume.shape, kernel.shape):
            # Convolution is commutative; order doesn't have any effect on output
            volume, kernel = kernel, volume

        print("volume shape: ", volume.shape)
        print("kernel shape: ", kernel.shape)

        out = self.fft_convolve(volume, kernel, mode=mode)

        return out.astype(np.result_type(volume, kernel), copy=False)

    def fft_convolve(self, in1, in2, mode="full", axes=None):
        """FFT convolve

        Parameters
        ----------
        in1
        in2
        mode
        axes

        """

        in1, in2, axes = _init_freq_conv_axes(in1, in2, mode, axes, sorted_axes=False)

        s1 = np.asarray(in1.shape)
        s2 = np.asarray(in2.shape)

        shape = tuple(s1 + s2 - 1)

        if not len(axes):
            return in1 * in2

        fft, ifft = self.get_right_fft(in1.ndim), self.get_right_ifft(in1.ndim)

        sp1 = fft(in1, shape)
        sp2 = fft(in2, shape)

        ret = ifft(sp1 * sp2, shape)
        ret = ret.numpy()

        return _apply_conv_mode(ret, s1, s2, mode, axes)

    def get_right_fft(self, ndim):
        if ndim == 1:
            return tf.signal.rfft
        elif ndim == 2:
            return tf.signal.rfft2d
        elif ndim == 3:
            return tf.signal.rfft3d
        else:
            raise Exception("Only convolutions up to 3-dimensions allowed!")

    def get_right_ifft(self, ndim):
        if ndim == 1:
            return tf.signal.irfft
        elif ndim == 2:
            return tf.signal.irfft2d
        elif ndim == 3:
            return tf.signal.irfft3d
        else:
            raise Exception("Only convolutions up to 3-dimensions allowed!")



import time

import numpy
import pytest
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import normalise, add_noise
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.cb import CBRegressor
from aydin.regression.lgbm import LGBMRegressor
from aydin.regression.linear import LinearRegressor
from aydin.regression.nn import NNRegressor
from aydin.regression.rf import RFRegressor
from aydin.regression.sv import SVRegressor


@pytest.mark.heavy
def test_it_fgr_linear():
    regressor = LinearRegressor()
    do_it_fgr_with_self_supervised(regressor, min_psnr=18, min_ssim=0.65)


def test_it_fgr_rf():
    regressor = RFRegressor(max_num_estimators=128)
    do_it_fgr_with_self_supervised(regressor, min_ssim=0.75)


@pytest.mark.heavy
def test_it_fgr_svr():
    regressor = SVRegressor()
    do_it_fgr_with_self_supervised(regressor, min_psnr=22, min_ssim=0.71)


def test_it_fgr_lgbm():
    regressor = LGBMRegressor(max_num_estimators=256)
    do_it_fgr_with_self_supervised(regressor, min_ssim=0.77)


@pytest.mark.heavy
def test_it_fgr_nn():
    regressor = NNRegressor(max_epochs=64)
    do_it_fgr_with_self_supervised(regressor, min_ssim=0.70)


def test_it_fgr_cb():
    regressor = CBRegressor(max_num_estimators=256, min_num_estimators=64)
    do_it_fgr_with_self_supervised(
        regressor, min_ssim=0.78
    )  # TODO: original was 0.85 investigate later


def test_it_fgr_cb_supervised():
    regressor = CBRegressor(max_num_estimators=256, min_num_estimators=64)
    do_it_fgr_with_supervised(regressor, min_ssim=0.77)


def do_it_fgr_with_self_supervised(regressor, min_psnr=22, min_ssim=0.75):
    """
    Test for self-supervised denoising using camera image with synthetic noise
    """

    image = normalise(camera().astype(numpy.float32))
    noisy = add_noise(image)

    generator = StandardFeatureGenerator(max_level=5, include_spatial_features=True)
    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    start = time.time()
    it.train(noisy, noisy)
    stop = time.time()
    print(f"####### Training: elapsed time:  {stop - start} sec")

    start = time.time()
    denoised = it.translate(noisy)
    stop = time.time()
    print(f"####### Inference: elapsed time:  {stop - start} sec")

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy

    # if the line below fails, then the parameters of the image the regressor have been broken.
    # do not change the number below, but instead, fix the problem -- most likely a parameter.

    assert psnr_denoised > min_psnr and ssim_denoised > min_ssim


def do_it_fgr_with_supervised(regressor, min_psnr=23, min_ssim=0.74):
    """
    Test for self-supervised denoising using camera image with synthetic noise
    """

    image = normalise(camera().astype(numpy.float32))
    noisy = add_noise(image)

    generator = StandardFeatureGenerator(include_spatial_features=False)
    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    start = time.time()
    it.train(noisy, image)
    stop = time.time()
    print(f"####### Training: elapsed time:  {stop - start} sec")

    start = time.time()
    denoised = it.translate(noisy)
    stop = time.time()
    print(f"####### Inference: elapsed time:  {stop - start} sec")

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(noisy, image)
    ssim_noisy = ssim(noisy, image)
    print("noisy", psnr_noisy, ssim_noisy)

    psnr_denoised = psnr(denoised, image)
    ssim_denoised = ssim(denoised, image)
    print("denoised", psnr_denoised, ssim_denoised)

    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy
    assert psnr_denoised > psnr_noisy and ssim_denoised > ssim_noisy

    # if the line below fails, then the parameters of the image the regressor have been broken.
    # do not change the number below, but instead, fix the problem -- most likely a parameter.

    assert psnr_denoised > min_psnr and ssim_denoised > min_ssim



import numpy
import torch
import torch.nn.functional as F
from scipy.signal import convolve2d

from aydin.nn.pytorch.it_ptcnn import PTCNNImageTranslator
from aydin.nn.pytorch.models.psf_convolution import PSFConvolutionLayer
from aydin.util.log.log import lprint


def to_numpy(tensor):
    return tensor.clone().detach().cpu().numpy()


class PTCNNDeconvolution(PTCNNImageTranslator):
    """
    Pytorch-based CNN image deconvolution
    """

    def __init__(self, psf_kernel=None, broaden_psf=1, **kwargs):
        """
        Constructs a CNN image translator using the pytorch deep learning library.

        :param normaliser_type: normaliser type
        :param balance_training_data: balance data ? (limits number training entries per target value histogram bin)
        :param monitor: monitor to track progress of training externally (used by UI)
        """
        super().__init__(**kwargs)

        for i in range(broaden_psf):
            psf_kernel = numpy.pad(psf_kernel, (1,), mode='constant', constant_values=0)
            psf_kernel = convolve2d(
                psf_kernel,
                numpy.array([[0.5, 1, 0.5], [1, 1, 1], [0.5, 1, 0.5]]) / 7,
                'same',
            )

        psf_kernel /= psf_kernel.sum()
        psf_kernel = psf_kernel.astype(numpy.float32)

        self.psf_kernel = psf_kernel
        self.psf_kernel_tensor = torch.from_numpy(
            self.psf_kernel[numpy.newaxis, numpy.newaxis, ...]
        ).to(self.device)

        self.psfconv = PSFConvolutionLayer(self.psf_kernel).to(self.device)

        self.enforce_blind_spot = False

        self.sharpening = 0.01

    def _train_loop(self, data_loader, optimizer, loss_function):
        try:
            self.model.kernel_continuity_regularisation = False
        except AttributeError:
            lprint("Cannot deactivate kernel continuity regularisation")

        super()._train_loop(data_loader, optimizer, loss_function)

    def _additional_losses(self, translated_image, forward_model_image):

        # non-negativity loss:
        non_negativity_loss = F.relu(-translated_image)
        loss = non_negativity_loss.mean()

        # Sharpen loss_deconvolution:
        if self.sharpening and self.sharpening > 0:
            sum_values = torch.sum(translated_image, (2, 3), keepdim=True)
            entropy = self.sharpening * torch.sqrt(1 + translated_image / sum_values)
            entropy_loss_value = entropy.mean()
            loss += entropy_loss_value

        return loss

    def _forward_model(self, input):
        return self.psfconv(input)



# flake8: noqa
import time

from skimage.exposure import rescale_intensity

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io import io
from aydin.io.datasets import examples_single
from aydin.it.classic import ImageDenoiserClassic
from aydin.it.classic_denoisers.butterworth import (
    calibrate_denoise_butterworth,
    denoise_butterworth,
)
from aydin.it.classic_denoisers.spectral import (
    calibrate_denoise_spectral,
    denoise_spectral,
)
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.padding import PaddingTransform
from aydin.it.transforms.range import RangeTransform
from aydin.regression.cb import CBRegressor
from aydin.util.log.log import Log


def demo():
    Log.enable_output = True

    image_path = examples_single.hyman_hela.get_path()
    image, metadata = io.imread(image_path)
    print(image.shape)
    image = rescale_intensity(image, in_range='image', out_range=(0, 1))
    # image = image[0:3]

    it = ImageDenoiserClassic(method="spectral")

    it.transforms_list.append(RangeTransform())
    it.transforms_list.append(PaddingTransform())

    batch_dims = (False, True, False, False)

    start = time.time()
    it.train(image, image, batch_axes=batch_dims)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    start = time.time()
    denoised = it.translate(image, batch_axes=batch_dims)
    stop = time.time()
    print(f"inference train: elapsed time:  {stop - start} ")

    print(image.shape)
    print(denoised.shape)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(denoised, name='denoised')


demo()



import napari
from qtpy.QtCore import Qt, QSize
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QPushButton,
    QTabWidget,
    QApplication,
    QStyle,
    QDockWidget,
)

from aydin.gui._qt.custom_widgets.activity_widget import ActivityWidget
from aydin.gui._qt.custom_widgets.overlay import Overlay
from aydin.gui._qt.custom_widgets.program_flow_diagram import QProgramFlowDiagramWidget
from aydin.gui._qt.job_runners.denoise_job_runner import DenoiseJobRunner
from aydin.gui.tabs.data_model import DataModel
from aydin.gui.tabs.qt.dimensions import DimensionsTab
from aydin.gui.tabs.qt.denoise import DenoiseTab
from aydin.gui.tabs.qt.files import FilesTab
from aydin.gui.tabs.qt.images import ImagesTab
from aydin.gui.tabs.qt.denoising_cropping import DenoisingCroppingTab
from aydin.gui.tabs.qt.processing import ProcessingTab
from aydin.gui.tabs.qt.summary import SummaryTab
from aydin.gui.resources.json_resource_loader import JSONResourceLoader
from aydin.gui.tabs.qt.training_cropping import TrainingCroppingTab
from aydin.io.utils import get_options_json_path
from aydin.util.misc.json import save_any_json


class MainPage(QWidget):
    """Start Page widget. It is the first widget welcomes user when
    GUI is started.

    Parameters
    ----------
    parent : object
    threadpool : QThreadPool

    """

    def __init__(self, parent, threadpool):
        super(QWidget, self).__init__(parent)
        self.parent = parent
        self.threadpool = threadpool

        self.setAcceptDrops(True)

        self.tooltips = JSONResourceLoader(resource_file_name="tooltips.json")

        self.tabs = {
            "Summary": SummaryTab(self),
            "File(s)": FilesTab(self),
            "Image(s)": ImagesTab(self),
            "Dimensions": DimensionsTab(self),
            "Training Crop": TrainingCroppingTab(self),
            "Denoising Crop": DenoisingCroppingTab(self),
            "Pre/Post-Processing": ProcessingTab(self),
            "Denoise": DenoiseTab(self),
        }

        self.data_model = DataModel(self)

        self.activity_dock = QDockWidget("Activity", self)

        # MainPage layout
        self.widget_layout = QVBoxLayout()
        self.widget_layout.setAlignment(Qt.AlignTop)

        # navbar
        self.navbar_layout = QHBoxLayout()
        self.navbar_layout.setAlignment(Qt.AlignCenter)
        self.navbar_layout.setContentsMargins(0, 0, 0, 0)
        self.navbar_layout.setSpacing(5)

        self.add_activity_dockable()

        self.navbar_layout_center = QHBoxLayout()

        self.flow_diagram_widget = QProgramFlowDiagramWidget(self)
        self.navbar_layout_center.addWidget(self.flow_diagram_widget)

        self.flow_diagram_widget.add_files_button.clicked.connect(
            self.tabs["File(s)"].openFileNamesDialog
        )
        self.flow_diagram_widget.add_files_button.setIcon(
            QApplication.style().standardIcon(QStyle.SP_DirIcon)
        )
        self.flow_diagram_widget.add_files_button.setToolTip(
            self.tooltips.json["main_page"]["open_file_button"]
        )

        self.flow_diagram_widget.files_button.clicked.connect(
            lambda: self.tabwidget.setCurrentIndex(1)
        )
        self.flow_diagram_widget.images_button.clicked.connect(
            lambda: self.tabwidget.setCurrentIndex(2)
        )
        self.flow_diagram_widget.dimensions_button.clicked.connect(
            lambda: self.tabwidget.setCurrentIndex(3)
        )
        self.flow_diagram_widget.training_crop_button.clicked.connect(
            lambda: self.tabwidget.setCurrentIndex(4)
        )
        self.flow_diagram_widget.inference_crop_button.clicked.connect(
            lambda: self.tabwidget.setCurrentIndex(5)
        )
        self.flow_diagram_widget.preprocess_button.clicked.connect(
            lambda: self.tabwidget.setCurrentIndex(6)
        )
        self.flow_diagram_widget.postprocess_button.clicked.connect(
            lambda: self.tabwidget.setCurrentIndex(6)
        )
        self.flow_diagram_widget.denoise_button.clicked.connect(
            lambda: self.tabwidget.setCurrentIndex(7)
        )

        self.start_button = QPushButton(
            "Start", icon=QApplication.style().standardIcon(QStyle.SP_MediaPlay)
        )
        self.start_button.setIconSize(QSize(30, 30))
        self.start_button.setMinimumHeight(30)
        self.start_button.setMinimumWidth(70)
        self.start_button.setStyleSheet("QPushButton {font-weight: bold;}")
        self.navbar_layout_center.addWidget(self.start_button)

        self.processing_job_runner = DenoiseJobRunner(
            self, self.threadpool, self.start_button
        )

        self.napari_button = QPushButton("View images")
        self.napari_button.setStyleSheet("QPushButton {font-weight: bold;}")
        self.napari_button.setMinimumHeight(30)
        self.napari_button.clicked.connect(self.open_images_with_napari)
        self.napari_button.setIcon(
            QApplication.style().standardIcon(QStyle.SP_CommandLink)
        )
        self.napari_button.setIconSize(QSize(30, 30))
        self.napari_button.setToolTip(self.tooltips.json["main_page"]["napari_button"])
        self.navbar_layout_center.addWidget(self.napari_button)

        self.navbar_layout_center.setAlignment(Qt.AlignCenter)
        self.navbar_layout.addLayout(self.navbar_layout_center)

        self.navbar_layout_right = QHBoxLayout()
        self.activity_button = QPushButton("Activity")
        self.activity_button.setStyleSheet("QPushButton {font-weight: bold;}")
        self.activity_button.setMinimumHeight(30)
        self.activity_button.clicked.connect(
            lambda: self.activity_dock.setHidden(not self.activity_dock.isHidden())
        )
        self.activity_button.setIcon(
            QApplication.style().standardIcon(QStyle.SP_ComputerIcon)
        )
        self.activity_button.setIconSize(QSize(30, 30))
        self.activity_button.setToolTip(
            self.tooltips.json["main_page"]["activity_button"]
        )
        self.navbar_layout_center.addWidget(self.activity_button)

        self.navbar_layout_right.setAlignment(Qt.AlignRight)
        self.navbar_layout.addLayout(self.navbar_layout_right)

        self.navbar_layout.setAlignment(Qt.AlignTop)
        self.widget_layout.addLayout(self.navbar_layout)

        # TabWidget
        self.tabwidget = QTabWidget(self)
        self.tabwidget.currentChanged.connect(self.onTabChange)
        for key, value in self.tabs.items():
            self.tabwidget.addTab(value, key)

        self.widget_layout.addWidget(self.tabwidget)

        self.overlay = Overlay(self)
        self.overlay.hide()

        # Set layout for the main page widget
        self.setLayout(self.widget_layout)

        self.tabs["Dimensions"].dimensions = None
        self.tabs["Training Crop"].images = []
        self.tabs["Denoising Crop"].images = []

    def onTabChange(self, current_tab_index):
        self.flow_diagram_widget.highlight_button(
            self.tabwidget.tabText(current_tab_index)
        )

    def enable_disable_a_tab(self, class_of_tab_to_disable, enable):
        for index, tab in enumerate(self.tabs.values()):
            if tab.__class__ == class_of_tab_to_disable:
                self.tabwidget.setTabEnabled(index, enable)
                return

    def resizeEvent(self, event):
        self.overlay.resize(event.size())
        event.accept()

    # The following three methods set up dragging and dropping for the app
    def dragEnterEvent(self, e):
        if e.mimeData().hasUrls:
            e.accept()
        else:
            e.ignore()

    def dragMoveEvent(self, e):
        if e.mimeData().hasUrls:
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, e):
        """Drop files directly onto the widget
        File locations are stored in fname

        Parameters
        ----------
        e

        """
        if e.mimeData().hasUrls:
            e.setDropAction(Qt.CopyAction)
            e.accept()
            fnames = []

            for url in e.mimeData().urls():
                fnames.append(str(url.toLocalFile()))

            self.tabwidget.setCurrentIndex(1)
            self.data_model.add_filepaths(fnames)
        else:
            e.ignore()

    def load_sample_image(self, sample):
        self.data_model.add_filepaths([sample.get_path()])
        self.tabwidget.setCurrentIndex(1)

    def handle_use_same_crop_state_changed(self):
        self.enable_disable_a_tab(
            self.tabs["Denoising Crop"].__class__,
            not self.tabs["Training Crop"].use_same_crop_checkbox.isChecked(),
        )

    def add_activity_dockable(self):
        self.activity_dock.setHidden(True)
        self.parent.addDockWidget(Qt.BottomDockWidgetArea, self.activity_dock)
        self.activity_widget = ActivityWidget(self)
        self.activity_dock.setWidget(self.activity_widget)

    def open_images_with_napari(self):
        training_images = self.tabs["Training Crop"].images
        inference_images = (
            self.tabs["Denoising Crop"].images
            if self.tabwidget.isTabEnabled(
                self.tabwidget.indexOf(self.tabs["Denoising Crop"])
            )
            else self.tabs["Training Crop"].images
        )

        viewer = napari.Viewer()
        viewer.show()

        if not self.tabwidget.isTabEnabled(
            self.tabwidget.indexOf(self.tabs["Denoising Crop"])
        ):
            # Training&Inference images
            for ind, image_array in enumerate(training_images):
                viewer.add_image(image_array, name=f"{ind}-training&inference_input")
        else:
            # Training images
            for ind, image_array in enumerate(training_images):
                viewer.add_image(image_array, name=f"{ind}-training_input")

            # Inference images
            for ind, image_array in enumerate(inference_images):
                viewer.add_image(image_array, name=f"{ind}-inference_input")

        # Result images
        for ind, image_array in enumerate(
            self.processing_job_runner.result_image_arrays()
        ):
            viewer.add_image(image_array, name=f"{ind}-denoised_output")

        viewer.grid.enabled = True

    def save_options_json(self, path=None):
        args_dict = self.tabs["Denoise"].lower_level_args

        if path is None:
            image_paths = [
                get_options_json_path(i[5]) for i in self.data_model.images_to_denoise
            ]
        else:
            image_paths = [path]

        for path in image_paths:
            save_any_json(args_dict, path)

    def filestab_changed(self):
        self.tabs["File(s)"].on_data_model_update()

    def imagestab_changed(self):
        self.tabs["Image(s)"].on_data_model_update()

    def dimensionstab_changed(self):
        self.tabs["Dimensions"].on_data_model_update()

    def croppingtabs_changed(self):
        self.tabs["Training Crop"].on_data_model_update()
        self.tabs["Denoising Crop"].on_data_model_update()



from pprint import pprint

import numpy

from aydin.util.blindspot.blindspot_transform import BlindSpotTransform


def test_blindspot_transform():

    blind_spots = [(0, -1), (0, 0), (0, 1)]
    bst = BlindSpotTransform(blind_spots)

    array = numpy.arange(12).reshape(2, 6)

    t_input_array, t_target_array = bst.transform(array, array)

    # We fake perfect inference here:
    t_translated_array = t_target_array

    it_array = bst.inverse_transform(t_translated_array)

    pprint(array)
    pprint(t_input_array)
    pprint(t_target_array)
    pprint(it_array)

    assert (it_array == array).all()



import numpy

from aydin.util.offcore.offcore import offcore_array


def test_offcore_array():

    array = offcore_array(shape=(1000, 1000), dtype=numpy.float32)

    vector = numpy.random.rand(1000).astype(numpy.float32)

    for i in range(1000):
        print(i)
        array[:, i] = vector

        assert array[:, i].sum() == vector.sum()

    del array



import torch
import torch.nn as nn
import torch.nn.functional as F

from aydin.nn.pytorch.models.convblock import ConvBlock


class Unet(nn.Module):
    def __init__(
        self,
        n_channel_in=1,
        n_channel_out=1,
        n_internal_channels=8,
        residual=True,
        down='avgpool',
        up='bilinear',
        activation='leakyrelu',
        norm=None,
        softmax=False,
    ):
        super(Unet, self).__init__()

        self.residual = residual
        self.softmax = softmax

        nic = n_internal_channels

        if down == 'maxpool':
            self.down1 = nn.MaxPool2d(kernel_size=2)
            self.down2 = nn.MaxPool2d(kernel_size=2)
            self.down3 = nn.MaxPool2d(kernel_size=2)
            self.down4 = nn.MaxPool2d(kernel_size=2)
        elif down == 'avgpool':
            self.down1 = nn.AvgPool2d(kernel_size=2)
            self.down2 = nn.AvgPool2d(kernel_size=2)
            self.down3 = nn.AvgPool2d(kernel_size=2)
            self.down4 = nn.AvgPool2d(kernel_size=2)
        elif down == 'convpool':
            self.down1 = nn.Conv2d(nic, nic, kernel_size=2, stride=2, groups=32)
            self.down2 = nn.Conv2d(nic * 2, nic * 2, kernel_size=2, stride=2, groups=64)
            self.down3 = nn.Conv2d(
                nic * 4, nic * 4, kernel_size=2, stride=2, groups=128
            )
            self.down4 = nn.Conv2d(
                nic * 8, nic * 8, kernel_size=2, stride=2, groups=256
            )

            self.down1.weight.data = 0.01 * self.down1.weight.data + 0.25
            self.down2.weight.data = 0.01 * self.down2.weight.data + 0.25
            self.down3.weight.data = 0.01 * self.down3.weight.data + 0.25
            self.down4.weight.data = 0.01 * self.down4.weight.data + 0.25

            self.down1.bias.data = 0.01 * self.down1.bias.data + 0
            self.down2.bias.data = 0.01 * self.down2.bias.data + 0
            self.down3.bias.data = 0.01 * self.down3.bias.data + 0
            self.down4.bias.data = 0.01 * self.down4.bias.data + 0

        if up == 'bilinear' or up == 'nearest':
            self.up1 = lambda x: nn.functional.interpolate(
                x, mode=up, scale_factor=2, align_corners=False
            )
            self.up2 = lambda x: nn.functional.interpolate(
                x, mode=up, scale_factor=2, align_corners=False
            )
            self.up3 = lambda x: nn.functional.interpolate(
                x, mode=up, scale_factor=2, align_corners=False
            )
            self.up4 = lambda x: nn.functional.interpolate(
                x, mode=up, scale_factor=2, align_corners=False
            )
        elif up == 'tconv':
            self.up1 = nn.ConvTranspose2d(
                nic * 8, nic * 8, kernel_size=2, stride=2, groups=nic * 8
            )
            self.up2 = nn.ConvTranspose2d(
                nic * 4, nic * 4, kernel_size=2, stride=2, groups=nic * 4
            )
            self.up3 = nn.ConvTranspose2d(
                nic * 2, nic * 2, kernel_size=2, stride=2, groups=nic * 2
            )
            self.up4 = nn.ConvTranspose2d(nic, nic, kernel_size=2, stride=2, groups=nic)

            self.up1.weight.data = 0.01 * self.up1.weight.data + 0.25
            self.up2.weight.data = 0.01 * self.up2.weight.data + 0.25
            self.up3.weight.data = 0.01 * self.up3.weight.data + 0.25
            self.up4.weight.data = 0.01 * self.up4.weight.data + 0.25

            self.up1.bias.data = 0.01 * self.up1.bias.data + 0
            self.up2.bias.data = 0.01 * self.up2.bias.data + 0
            self.up3.bias.data = 0.01 * self.up3.bias.data + 0
            self.up4.bias.data = 0.01 * self.up4.bias.data + 0

        self.conv1 = ConvBlock(
            n_channel_in, nic, residual=residual, activation=activation, norm=norm
        )
        self.conv2 = ConvBlock(
            nic, nic * 2, residual=residual, activation=activation, norm=norm
        )
        self.conv3 = ConvBlock(
            nic * 2, nic * 4, residual=residual, activation=activation, norm=norm
        )
        self.conv4 = ConvBlock(
            nic * 4, nic * 8, residual=residual, activation=activation, norm=norm
        )

        self.conv5 = ConvBlock(
            nic * 8, nic * 8, residual=residual, activation=activation, norm=norm
        )

        self.conv6 = ConvBlock(
            2 * nic * 8, nic * 4, residual=residual, activation=activation, norm=norm
        )
        self.conv7 = ConvBlock(
            2 * nic * 4, nic * 2, residual=residual, activation=activation, norm=norm
        )
        self.conv8 = ConvBlock(
            2 * nic * 2, nic, residual=residual, activation=activation, norm=norm
        )
        self.conv9 = ConvBlock(
            2 * nic, n_channel_out, residual=residual, activation=activation, norm=norm
        )

        if self.residual:
            self.convres = ConvBlock(
                n_channel_in,
                n_channel_out,
                residual=residual,
                activation=activation,
                norm=norm,
            )

    def forward(self, x):
        c0 = x
        c1 = self.conv1(x)
        x = self.down1(c1)
        c2 = self.conv2(x)
        x = self.down2(c2)
        c3 = self.conv3(x)
        x = self.down3(c3)
        c4 = self.conv4(x)
        x = self.down4(c4)
        if self.softmax:
            x = F.softmax(x, dim=1)
        x = self.conv5(x)
        x = self.up1(x)
        # print("shapes: c0:%sx:%s c4:%s " % (c0.shape,x.shape,c4.shape))
        if self.softmax:
            x = F.softmax(x, dim=1)
        x = torch.cat([x, c4], 1)  # x[:,0:128]*x[:,128:256],
        x = self.conv6(x)
        x = self.up2(x)
        if self.softmax:
            x = F.softmax(x, dim=1)
        x = torch.cat([x, c3], 1)  # x[:,0:64]*x[:,64:128],
        x = self.conv7(x)
        x = self.up3(x)
        if self.softmax:
            x = F.softmax(x, dim=1)
        x = torch.cat([x, c2], 1)  # x[:,0:32]*x[:,32:64],
        x = self.conv8(x)
        x = self.up4(x)
        if self.softmax:
            x = F.softmax(x, dim=1)
        x = torch.cat([x, c1], 1)  # x[:,0:16]*x[:,16:32],
        x = self.conv9(x)
        if self.residual:
            x = torch.add(x, self.convres(c0))

        return x



# flake8: noqa

from aydin.io.datasets import examples_single
from aydin.it.classic_denoisers.dictionary_learned import (
    calibrate_denoise_dictionary_learned,
)

from aydin.util.log.log import Log


def demo_dict_learn_hela(display=True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True
    Log.set_log_max_depth(5)

    noisy = examples_single.hyman_hela.get_array()[10]

    function, parameters, _ = calibrate_denoise_dictionary_learned(
        noisy, dictionary_type='learned', display_dictionary=display
    )
    denoised = function(noisy, **parameters)

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(noisy, name='noisy')
            viewer.add_image(denoised, name='denoised')


if __name__ == "__main__":
    demo_dict_learn_hela()



from numpy.random.mtrand import randint

from aydin.analysis.empirical_noise_model import (
    distill_noise_model,
    sample_noise_from_model,
)
from aydin.io.datasets import camera


def demo_noise_model():
    clean_image = camera()

    noise = randint(0, 60, size=clean_image.size).reshape(*clean_image.shape)
    noisy_image = clean_image + noise

    noise_model = distill_noise_model(clean_image, noisy_image)

    noisy_image_sampled = sample_noise_from_model(clean_image, noise_model)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(clean_image, name='clean_image')
        viewer.add_image(noisy_image, name='noisy_image')
        viewer.add_image(noisy_image_sampled, name='noisy_image_sampled')
        viewer.add_image(noise_model, name='noise_model')


demo_noise_model()



import torch
import torch.nn as nn
import torch.nn.functional as F


class VeryFlatNet(nn.Module):
    def __init__(self, num_channels=128, kernel_size=9):
        super(VeryFlatNet, self).__init__()

        self.num_channels = num_channels
        print("num_channels =%d" % num_channels)

        padding = int((kernel_size - 1) / 2)

        self.convfeatures = nn.Conv2d(
            1,
            num_channels,
            groups=1,
            kernel_size=kernel_size,
            padding=padding,
            stride=1,
        )

        channels = 1 + num_channels

        self.convp0 = nn.Conv2d(channels, channels // 2, kernel_size=1, padding=0)
        channels = channels // 2

        self.convp1 = nn.Conv2d(channels, channels // 2, kernel_size=1, padding=0)
        channels = channels // 2

        self.convp2 = nn.Conv2d(channels, channels // 2, kernel_size=1, padding=0)
        channels = channels // 2

        self.convpf = nn.Conv2d(channels, 1, kernel_size=1, padding=0)

    def set_weights(self, weights, bias=0):
        device = next(self.parameters()).device
        with torch.no_grad():
            length = weights.shape[0]
            print(length)
            self.convfeatures._parameters['weight'][0:length] = torch.from_numpy(
                weights
            ).to(device)
            # self.convfeatures._parameters['bias']   = bias*torch.ones([self.num_channels], dtype=torch.float32, device=device)

    def lastparameters(self):
        from itertools import chain

        return chain(
            self.convp0.parameters(),
            self.convp1.parameters(),
            self.convp2.parameters(),
            self.convpf.parameters(),
        )

    def verylastparameters(self):
        return self.convp5.parameters()

    def forward(self, x):
        y = self.convfeatures(x)

        features = F.relu(torch.cat((x, y), 1))
        features = F.relu(self.convp0(features))
        features = F.relu(self.convp1(features))
        features = F.relu(self.convp2(features))
        prediction = self.convpf(features)

        return prediction



# flake8: noqa
import math
import os
import time

import numpy
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.analysis.image_metrics import mutual_information
from aydin.io.datasets import (
    normalise,
    add_noise,
    newyork,
    characters,
    lizard,
    scafoldings,
    pollen,
    andromeda,
    fibsem,
    dots,
)
from aydin.nn.pytorch.it_ptcnn import PTCNNImageTranslator
from aydin.util.log.log import Log


def demo(image, name, do_add_noise=True):

    Log.enable_output = True
    # image = image[0:1024, 0:1024]

    image = normalise(image.astype(numpy.float32))

    noisy = add_noise(image) if do_add_noise else image
    # noisy_image = image

    it = PTCNNImageTranslator(
        max_epochs=1024,
        patience=math.inf,
        learning_rate=0.01,
        normaliser_type='identity',
    )

    start = time.time()
    it.train(noisy, noisy)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    # in case of batching we have to do this:
    start = time.time()
    denoised = it.translate(noisy)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    print(denoised.shape)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    # denoised = numpy.clip(denoised, 0, 1)

    psnr_noisy = psnr(image, noisy)
    mi_noisy = mutual_information(image, noisy)
    ssim_noisy = ssim(image, noisy)
    psnr_denoised = psnr(image, denoised)
    mi_denoised = mutual_information(image, denoised)
    ssim_denoised = ssim(image, denoised)

    print("noisy       :", psnr_noisy, mi_noisy, ssim_noisy)
    print("denoised_inf:", psnr_denoised, mi_denoised, ssim_denoised)

    # visualise weights with napari:
    # it.visualise_weights()

    import matplotlib.pyplot as plt

    plt.figure(figsize=(2.7 * 5, 5), dpi=300)
    plt.subplot(1, 3, 1)
    plt.imshow(normalise(noisy), cmap='gray')
    plt.axis('off')
    plt.title(f'Noisy \nPSNR: {psnr_noisy:.3f}, SSIM: {ssim_noisy:.3f}')
    plt.subplot(1, 3, 2)
    plt.imshow(normalise(denoised), cmap='gray')
    plt.axis('off')
    plt.title(f'Denoised \nPSNR: {psnr_denoised:.3f}, SSIM: {ssim_denoised:.3f}')
    plt.subplot(1, 3, 3)
    plt.imshow(normalise(image), cmap='gray')
    plt.axis('off')
    plt.title('Original')
    plt.subplots_adjust(left=0.01, right=0.99, top=0.95, bottom=0.01, hspace=0.1)
    os.makedirs("../../demo/demo_results", exist_ok=True)
    plt.savefig(f'../../demo/demo_results/n2s_jidcnet_2D_{name}.png')

    it.visualise_weights()

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(normalise(image), name='image')
        viewer.add_image(normalise(noisy), name='noisy')
        viewer.add_image(normalise(denoised), name='denoised')

    return ssim(denoised, image)


camera_image = camera()
demo(camera_image, "camera")
dots_image = dots()
demo(dots_image, "dots")
# lizard_image = lizard()
# demo(lizard_image, "lizard")
# pollen_image = pollen()
# demo(pollen_image, "pollen")
# newyork_image = newyork()
# demo(newyork_image, "newyork")
# characters_image = characters()
# demo(characters_image, "characters")
# scafoldings_image = scafoldings()
# demo(scafoldings_image, "scafoldings")
# andromeda_image = andromeda()
# demo(andromeda_image, "andromeda")
# fibsem_image = fibsem()
# demo(fibsem_image, "fibsem", False)



import numpy
from aydin.util.psf.microscope_psf import MicroscopePSF


class SimpleMicroscopePSF(MicroscopePSF):
    """
    Simple Microscope PSF: no cover-slip, i.e. Richards & Wolf as a degenerate case of Gibson & Lani
    """

    def __init__(self, M=16, NA=0.8, n=1.33, wd=3000):
        """ """
        super().__init__()

        # Microscope parameters.
        self.parameters["M"] = M  # magnification
        self.parameters["NA"] = NA  # numerical aperture
        self.parameters["ni0"] = n
        self.parameters["ni"] = n
        self.parameters["ns"] = n
        self.parameters["ti0"] = wd

    def generate_xyz_psf(self, dxy, dz, xy_size, z_size):
        """
        Generates a 3D PSF array.

        :param dxy: voxel dimension along xy (microns)
        :param dz: voxel dimension along z (microns)
        :param xy_size: size of PSF kernel along x and y (odd integer)
        :param z_size: size of PSF kernel along z (odd integer)

        """
        lz = (z_size) * dz
        z_offset = -(lz - 2 * dz) / 2
        pz = numpy.arange(0, lz, dz)

        # gLXYZParticleScan(self, dxy, xy_size, pz, normalize = True, wvl = 0.6, zd = None, zv = 0.0):
        psf_xyz_array = self.gLXYZParticleScan(
            dxy=dxy, xy_size=xy_size, pz=pz, zv=z_offset
        )
        print(psf_xyz_array.shape)

        return psf_xyz_array



import numpy

from aydin.features.extensible_features import ExtensibleFeatureGenerator
from aydin.features.groups.dct import DCTFeatures
from aydin.features.groups.median import MedianFeatures
from aydin.features.groups.random import RandomFeatures
from aydin.features.groups.spatial import SpatialFeatures
from aydin.features.groups.uniform import UniformFeatures
from aydin.util.log.log import lprint


class StandardFeatureGenerator(ExtensibleFeatureGenerator):
    """Standard Feature Generator"""

    def __init__(
        self,
        kernel_widths=None,
        kernel_scales=None,
        kernel_shapes=None,
        min_level: int = 0,
        max_level: int = 13,
        scale_one_width: int = 3,
        include_scale_one: bool = False,
        include_fine_features: bool = True,
        include_corner_features: bool = False,
        include_line_features: bool = False,
        decimate_large_scale_features: bool = True,
        extend_large_scale_features: bool = False,
        include_spatial_features: bool = False,
        spatial_features_coarsening: int = 2,
        num_sinusoidal_features: int = 0,
        include_median_features: bool = False,
        include_dct_features: bool = False,
        dct_max_freq: float = 0.5,
        include_random_conv_features: bool = False,
        dtype: numpy.dtype = numpy.float32,
    ):
        """Constructs a standard feature generator.

        Parameters
        ----------
        kernel_widths : numpy.typing.ArrayLike
            ArrayLike of kernel widths.
        kernel_scales : numpy.typing.ArrayLike
            ArrayLike of kernel scales.
        kernel_shapes : numpy.typing.ArrayLike
            ArrayLike of kernel shapes.
        min_level : int
            Minimum level of features to include
        max_level : int
            Maximum level of features to include
        scale_one_width : int
            Width of the scale one features
        include_scale_one : bool
            True or False, argument to set decision
            on inclusion of scale-one-features
        include_fine_features : bool
            True or False, argument to set decision
            on inclusion of fine features
        include_corner_features : bool
            True or False, argument to set decision
            on inclusion of corner features
        include_line_features : bool
            True or False, argument to set decision
            on inclusion of line features
        decimate_large_scale_features : bool
            True or False, argument to set decision
            on decimation of large scale features
        extend_large_scale_features : bool
            True or False, argument to set decision
            on extension of large scale features.
        include_spatial_features : bool
            True or False, argument to set decision
            on inclusion of spatial features
        spatial_features_coarsening : int
            Degree of coarsening to apply on spatial features
            to prevent identification of individual pixel values
        num_sinusoidal_features : int
            Number of sinusoidal features
        include_median_features : bool
            True or False, argument to set decision
            on inclusion of median features
        include_dct_features : bool
            True or False, argument to set decision
            on inclusion of dct features
        dct_max_freq : float
            Maximum included frequency during
            dct features computation.
        include_random_conv_features : bool
            True or False, argument to set decision
            on inclusion of random convolutional features
        dtype
            Datatype of the features
        """
        super().__init__()

        self.dtype = dtype
        lprint(f"Features will be computed using dtype: {dtype}")

        uniform = UniformFeatures(
            kernel_widths=kernel_widths,
            kernel_scales=kernel_scales,
            kernel_shapes=kernel_shapes,
            min_level=min_level,
            max_level=max_level,
            include_scale_one=include_scale_one,
            include_fine_features=include_fine_features,
            include_corner_features=include_corner_features,
            include_line_features=include_line_features,
            decimate_large_scale_features=decimate_large_scale_features,
            extend_large_scale_features=extend_large_scale_features,
            scale_one_width=scale_one_width,
            dtype=dtype,
        )
        self.add_feature_group(uniform)

        if include_spatial_features:
            spatial = SpatialFeatures(coarsening=spatial_features_coarsening)
            self.add_feature_group(spatial)

        if num_sinusoidal_features > 0:
            periods = list([1 / 2 ** i for i in range(num_sinusoidal_features)])
            for period in periods:
                spatial = SpatialFeatures(period=period)
                self.add_feature_group(spatial)

        if include_median_features:
            radii = [1, 2, 3]
            medians = MedianFeatures(radii=radii)
            self.add_feature_group(medians)

        if include_dct_features:
            dct3 = DCTFeatures(size=3, max_freq=dct_max_freq, power=1)
            dct5 = DCTFeatures(size=5, max_freq=dct_max_freq, power=1)
            dct7 = DCTFeatures(size=7, max_freq=dct_max_freq, power=1)
            self.add_feature_group(dct3)
            self.add_feature_group(dct5)
            self.add_feature_group(dct7)

        if include_random_conv_features:
            rnd3 = RandomFeatures(size=3, num_features=3)
            rnd5 = RandomFeatures(size=5, num_features=5)
            rnd7 = RandomFeatures(size=7, num_features=7)
            self.add_feature_group(rnd3)
            self.add_feature_group(rnd5)
            self.add_feature_group(rnd7)



# flake8: noqa
import numpy as np
from skimage.data import camera

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import (
    normalise,
    add_noise,
    dots,
    lizard,
    pollen,
    characters,
    cropped_newyork,
)
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.padding import PaddingTransform
from aydin.it.transforms.range import RangeTransform
from aydin.regression.cb import CBRegressor
from aydin.util.log.log import Log
from aydin.util.nsi.nearest_self_image import nearest_self_image


def demo_nearest_self_image(image, display: bool = True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True
    Log.set_log_max_depth(5)

    image = normalise(image.astype(np.float32))
    noisy = add_noise(image)

    self_image = nearest_self_image(image)
    self_noisy = nearest_self_image(noisy)

    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=True,
        include_fine_features=True,
        include_spatial_features=True,
    )

    regressor = CBRegressor(patience=32, gpu=True, min_num_estimators=1024)
    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    it.transforms_list.append(RangeTransform())
    it.transforms_list.append(PaddingTransform())

    it.train(self_noisy, noisy, jinv=False)

    denoised = it.translate(noisy)

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(image, name='image')
            viewer.add_image(noisy, name='noisy')
            viewer.add_image(self_image, name='self_image')
            viewer.add_image(self_noisy, name='self_noisy')
            viewer.add_image(denoised, name='denoised')


if __name__ == "__main__":
    newyork_image = cropped_newyork()  # [0:64, 0:64]
    demo_nearest_self_image(newyork_image)
    camera_image = camera()
    demo_nearest_self_image(camera_image)
    characters_image = characters()
    demo_nearest_self_image(characters_image)
    pollen_image = pollen()
    demo_nearest_self_image(pollen_image)
    lizard_image = lizard()
    demo_nearest_self_image(lizard_image)
    dots_image = dots()
    demo_nearest_self_image(dots_image)



import importlib
import inspect
import pkgutil

from qtpy.QtWidgets import QTabWidget

from aydin.gui._qt.transforms_tab_item import TransformsTabItem
from aydin.it import transforms


class TransformsTabWidget(QTabWidget):
    def __init__(self, parent):
        super(TransformsTabWidget, self).__init__(parent)

        self.parent = parent
        self.list_of_item_widgets = []

        for module in [
            x
            for x in pkgutil.iter_modules(tuple(transforms.__path__))
            if not x.ispkg and x.name != 'base'
        ]:
            name = module.name  # name is filename

            response = importlib.import_module(transforms.__name__ + '.' + module.name)
            elem = [
                x for x in dir(response) if module.name.replace('_', '') in x.lower()
            ][
                0
            ]  # class name

            class_itself = response.__getattribute__(elem)
            fullargspec = inspect.getfullargspec(class_itself.__init__)
            # print(fullargspec)

            widget = TransformsTabItem(
                self,
                name=class_itself.preprocess_description,
                arg_names=fullargspec.args[1:],
                arg_defaults=fullargspec.defaults,
                arg_annotations=fullargspec.annotations,
                transform_class=class_itself,
            )
            self.list_of_item_widgets.append(widget)

            self.addTab(widget, name.replace("_", " "))

            if name in ["range", "padding", "vst"]:
                widget.preprocess_checkbox.setChecked(True)

    def clear_the_list(self):
        self.clear()



from aydin.features.standard_features import StandardFeatureGenerator
from aydin.it.deconvolution.base_deconv import ImageTranslatorDeconvBase
from aydin.it.deconvolution.lr_deconv_scipy import ImageTranslatorLRDeconvScipy
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.cb import CBRegressor


class ImageTranslatorLearnedLRDeconv(ImageTranslatorDeconvBase):
    """Learned Lucy Richardson Deconvolution

    Idea: Instead of just applying Lucy_Richardson deconvolution, we learn the to translate from the input image to the LR deconvolved image,
    this let's us run LR for a high number of iterations and control the noise by learning to deconvolve.
    """

    def __init__(
        self,
        psf_kernel,
        *args,
        deconvolution_class=ImageTranslatorLRDeconvScipy,
        deconvolving_translator=None,
        denoising_translator=None,
        feature_generator_class=StandardFeatureGenerator,
        regressor_class=CBRegressor,
        **kwargs,
    ):
        """Constructs a Lucy Richardson deconvolution image translator.

        Parameters
        ----------
        psf_kernel
            2D or 3D kernel, dimensions should be odd numbers and numbers sum to 1
        args
        deconvolution_class
        deconvolving_translator
        denoising_translator
        feature_generator_class
        regressor_class
        kwargs
        """
        super().__init__(psf_kernel=psf_kernel)

        if deconvolving_translator is None:
            generator = feature_generator_class(
                include_spatial_features=True,
                include_scale_one=True,
                include_corner_features=True,
                include_fine_features=True,
            )
            regressor = regressor_class()

            deconvolving_translator = ImageTranslatorFGR(
                feature_generator=generator, regressor=regressor
            )

        if denoising_translator is None:
            generator = feature_generator_class(
                include_spatial_features=True,
                include_scale_one=True,
                include_corner_features=True,
                include_fine_features=True,
            )
            regressor = regressor_class()

            denoising_translator = ImageTranslatorFGR(
                feature_generator=generator, regressor=regressor
            )

        self.deconvolution_translator = deconvolution_class(psf_kernel, *args, **kwargs)
        self.learned_deconvolving_translator = deconvolving_translator
        self.learned_denoising_translator = denoising_translator

    def get_receptive_field_radius(self, ndim):
        """Returns the receptive field radius

        Parameters
        ----------
        ndim : int

        Returns
        -------
        int
            Receptive field radius

        """
        return self.deconvolution_translator.get_receptive_field_radius(ndim)

    def _train(self, *args, **kwargs):
        self.deconvolution_translator._train(*args, **kwargs)

    def _translate(self, *args, **kwargs):
        self.deconvolution_translator._translate(*args, **kwargs)

    def train(
        self,
        input_image,
        target_image=None,
        batch_axes=None,
        channel_axes=None,
        train_valid_ratio=0.1,
        callback_period=3,
        jinv=True,
    ):
        """Train method

        Parameters
        ----------
        input_image
        target_image
        batch_axes
        channel_axes
        train_valid_ratio
        callback_period
        jinv

        """

        self.learned_denoising_translator.train(target_image)
        denoised_target_image = self.learned_denoising_translator.translate(
            target_image
        )

        deconvolved_denoised_target_image = self.deconvolve(
            denoised_target_image,
            batch_axes=batch_axes,
            train_valid_ratio=train_valid_ratio,
            callback_period=callback_period,
            force_jinv=jinv,
        )

        self.learned_deconvolving_translator.train(
            input_image, deconvolved_denoised_target_image, jinv=False
        )

    def deconvolve(
        self,
        image,
        batch_axes=None,
        train_valid_ratio=0.1,
        callback_period=3,
        force_jinv=None,
    ):
        """Deconvolves the given image

        Parameters
        ----------
        image
        batch_axes
        train_valid_ratio
        callback_period
        force_jinv

        Returns
        -------
        numpy.typing.ArrayLike
            Deconvolved image

        """

        self.deconvolution_translator.train(
            image,
            batch_axes=batch_axes,
            train_valid_ratio=train_valid_ratio,
            callback_period=callback_period,
            jinv=force_jinv,
        )

        deconvolved_image = self.deconvolution_translator.translate(image)
        return deconvolved_image

    def translate(
        self,
        input_image,
        translated_image=None,
        batch_axes=None,
        channel_axes=None,
        tile_size=None,
        min_margin=8,
        max_margin=None,
        denormalise_values=True,
        leave_as_float=False,
        clip=True,
    ):
        """Translates the given image

        Parameters
        ----------
        input_image
        translated_image
        batch_axes
        channel_axes
        tile_size
        min_margin
        max_margin
        denormalise_values
        leave_as_float
        clip

        Returns
        -------
        numpy.typing.ArrayLike
            Translated image

        """
        return self.learned_deconvolving_translator.translate(
            input_image, translated_image, batch_axes, tile_size, max_margin
        )
